import copy, random, json, time, os, shutil, logging

from flask import (
    Flask,
    Blueprint,
    render_template,
    request,
    redirect,
    url_for,
    jsonify,
    flash,
)
from flask_babel import lazy_gettext as _
from flask_login import login_required, current_user
from flask import current_app as app
from ..rpc import get_default_datadir
from ..node import Node
from ..specter_error import ExtProcTimeoutException
from ..util.shell import get_last_lines_from_file

logger = logging.getLogger(__name__)

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
nodes_endpoint = Blueprint("nodes_endpoint", __name__)


@nodes_endpoint.route(
    "new_node/", defaults={"node_alias": None}, methods=["GET", "POST"]
)
@nodes_endpoint.route("node/<node_alias>/", methods=["GET", "POST"])
@login_required
def node_settings(node_alias):
    if node_alias:
        try:
            node = app.specter.node_manager.get_by_alias(node_alias)
            if not node.external_node:
                return redirect(
                    url_for(
                        "nodes_endpoint.internal_node_settings",
                        node_alias=node.alias,
                    )
                )
        except:
            return render_template(
                "base.jinja", error=_("Node not found"), specter=app.specter, rand=rand
            )
    else:
        node = Node.from_json(
            {
                "name": "New Node",
                "autodetect": True,
                "datadir": get_default_datadir(),
                "user": "",
                "password": "",
                "port": 8332,
                "host": "localhost",
                "protocol": "http",
                "external_node": True,
            },
            app.specter.node_manager,
        )

    if not current_user.is_admin:
        flash(_("Only an admin is allowed to access this page"), "error")
        return redirect("")
    # The node might have been down but is now up again
    # (and the checker did not realized yet) and the user clicked "Configure Node"
    if node.rpc is None and node_alias:
        node.update_rpc()

    test = None
    if request.method == "POST":
        action = request.form["action"]

        if action != "rename":
            autodetect = "autodetect" in request.form
            if autodetect:
                datadir = request.form["datadir"]
            else:
                datadir = ""
            user = request.form["username"]
            password = request.form["password"]
            port = request.form["port"]
            host = request.form["host"].rstrip("/")
            # protocol://host
            if "://" in host:
                arr = host.split("://")
                protocol = arr[0]
                host = arr[1]
            else:
                protocol = "http"

            if not node_alias:
                node.name = request.form["name"]

        if action == "rename":
            node_name = request.form["newtitle"]
            if not node_name:
                flash(_("Node name cannot be empty"), "error")
            elif node_name == node.name:
                pass
            elif node_name in app.specter.device_manager.devices_names:
                flash(_("Node with this name already exists"), "error")
            else:
                node.rename(node_name)
        elif action == "forget":
            if not node_alias:
                flash(_("Failed to deleted node. Node isn't saved"), "error")
            elif len(app.specter.node_manager.nodes) > 1:
                app.specter.node_manager.delete_node(node, app.specter)
                flash(_("Node deleted successfully"))
                return redirect(
                    url_for(
                        "nodes_endpoint.node_settings",
                        node_alias=app.specter.node.alias,
                    )
                )
            else:
                flash(
                    _(
                        "Failed to delete node. Specter must have at least one node configured"
                    ),
                    "error",
                )
        elif action == "test":
            # If this is failing, the test_rpc-method needs improvement
            # Don't wrap this into a try/except otherwise the feedback
            # of what's wrong to the user gets broken
            node = Node(
                node.name,
                node.alias,
                autodetect,
                datadir,
                user,
                password,
                port,
                host,
                protocol,
                node.external_node,
                node.fullpath,
                "BTC",
                node.manager,
            )
            test = node.test_rpc()

            if "tests" in test:
                # If any test has failed, we notify the user that the test has not passed
                if not test["tests"] or False in list(test["tests"].values()):
                    flash(_("Test failed: {}").format(test["err"]), "error")
                else:
                    flash(_("Test passed"), "info")
            elif "err" in test:
                flash(_("Test failed: {}").format(test["err"]), "error")
        elif action == "save":
            if not node_alias:
                if node.name in app.specter.node_manager.nodes:
                    flash(
                        _(
                            "Node with this name already exists, please choose a different name."
                        ),
                        "error",
                    )
                    return render_template(
                        "node/node_settings.jinja",
                        node=node,
                        node_alias=node_alias,
                        test=test,
                        specter=app.specter,
                        rand=rand,
                    )
                node = app.specter.node_manager.add_node(
                    "BTC",
                    node.name,
                    autodetect,
                    datadir,
                    user,
                    password,
                    port,
                    host,
                    protocol,
                    node.external_node,
                )
                app.specter.update_active_node(node.alias)
                return redirect(
                    url_for("nodes_endpoint.node_settings", node_alias=node.alias)
                )

            success = node.update_rpc(
                autodetect=autodetect,
                datadir=datadir,
                user=user,
                password=password,
                port=port,
                host=host,
                protocol=protocol,
            )
            if not success:
                flash(_("Failed connecting to the node"), "error")
            if app.specter.active_node_alias == node.alias:
                app.specter.check()

    return render_template(
        "node/node_settings.jinja",
        node=node,
        node_alias=node_alias,
        test=test,
        specter=app.specter,
        rand=rand,
    )


@nodes_endpoint.route("specter_node/<node_alias>/", methods=["GET", "POST"])
@login_required
def internal_node_settings(node_alias):
    err = None
    if node_alias:
        try:
            node = app.specter.node_manager.get_by_alias(node_alias)
            if node.external_node:
                return redirect(
                    url_for(
                        "nodes_endpoint.node_settings",
                        node_alias=node.alias,
                    )
                )
        except:
            return render_template(
                "base.jinja", error=_("Node not found"), specter=app.specter, rand=rand
            )
    else:
        # TODO: Allow internal node setup here?
        return redirect(
            url_for(
                "nodes_endpoint.internal_node_settings",
                node_alias=node.alias,
            )
        )

    if not current_user.is_admin:
        flash(_("Only an admin is allowed to access this page"), "error")
        return redirect("")
    # The node might have been down but is now up again
    # (and the checker did not realized yet) and the user clicked "Configure Node"
    if node.rpc is None or not node.is_bitcoind_running():
        node.update_rpc()

    if request.method == "POST":
        action = request.form["action"]

        if action == "rename":
            node_name = request.form["newtitle"]
            if not node_name:
                flash(_("Node name must not be empty"), "error")
            elif node_name == node.name:
                pass
            elif node_name in app.specter.device_manager.devices_names:
                flash("Node with this name already exists", "error")
            else:
                node.rename(node_name)
        elif action == "forget":
            if not node_alias:
                flash(_("Failed to deleted node. Node isn't saved"), "error")
            elif len(app.specter.node_manager.nodes) > 1:
                node.stop()
                app.specter.node_manager.delete_node(node, app.specter)
                if bool(request.form.get("remove_datadir", False)):
                    shutil.rmtree(os.path.expanduser(node.datadir), ignore_errors=True)
                flash(_("Node deleted successfully"))
                return redirect(
                    url_for(
                        "nodes_endpoint.node_settings",
                        node_alias=app.specter.node.alias,
                    )
                )
            else:
                flash(
                    _(
                        "Failed to delete node. Specter must have at least one node configured"
                    ),
                    "error",
                )
        elif action == "stopbitcoind":
            try:
                node.stop()
                time.sleep(5)
                flash(_("Specter successfully stopped Bitcoin Core"))
            except Exception as e:
                try:
                    logger.exception(e)
                    flash(_("Stopping Bitcoin Core, this might take a few moments."))
                    node.rpc.stop()
                except Exception as ne:
                    logger.exception(ne)
                    flash(_("Failed to stop Bitcoin Core {}").format(ne), "error")
        elif action == "startbitcoind":
            if node.start(timeout=120):
                flash(_("Specter has started Bitcoin Core"))
            else:
                flash(_("Specter failed to start the node"), "error")
        elif action == "uninstall_bitcoind":
            try:
                node.stop()
                shutil.rmtree(
                    os.path.join(app.specter.data_folder, "bitcoin-binaries"),
                    ignore_errors=True,
                )
                if bool(request.form.get("remove_datadir", False)):
                    shutil.rmtree(os.path.expanduser(node.datadir), ignore_errors=True)
                flash(_("Bitcoin Core successfully uninstalled"))
                app.specter.node_manager.delete_node(node, app.specter)
                return redirect(
                    url_for(
                        "nodes_endpoint.node_settings",
                        node_alias=app.specter.node.alias,
                    )
                )
            except Exception as e:
                flash(_("Failed to remove Bitcoin Core, error: {}").format(e), "error")
        elif action == "upgrade_bitcoind":
            if node.version != app.config["INTERNAL_BITCOIND_VERSION"]:
                try:
                    app.specter.node_manager.update_bitcoind_version(
                        app.specter, app.config["INTERNAL_BITCOIND_VERSION"]
                    )
                except Exception as e:
                    flash(
                        _("Failed to upgrade Bitcoin Core version, error: {}").format(
                            e
                        ),
                        "error",
                    )
            else:
                flash(_("Bitcoin Core version is already up to date"))
    return render_template(
        "node/internal_node_settings.jinja",
        node=node,
        latest_bitcoind=app.config["INTERNAL_BITCOIND_VERSION"],
        node_alias=node_alias,
        specter=app.specter,
        rand=rand,
    )


@nodes_endpoint.route("/internal_node_logs/<node_alias>/", methods=["GET"])
@login_required
def internal_node_logs(node_alias):
    node = app.specter.node_manager.get_by_alias(node_alias)
    logfile_location = os.path.join(node.datadir, "debug.log")
    return render_template(
        "node/internal_node_logs.jinja",
        node_alias=node_alias,
        specter=app.specter,
        loglines="".join(get_last_lines_from_file(logfile_location)),
    )


@nodes_endpoint.route("switch_node/", methods=["POST"])
@login_required
def switch_node():
    node_alias = request.form["node_alias"]
    app.specter.update_active_node(node_alias)
    return redirect(url_for("nodes_endpoint.node_settings", node_alias=node_alias))



import logging
from json.decoder import JSONDecodeError

import requests
from requests.exceptions import ConnectionError, HTTPError
from urllib3.exceptions import NewConnectionError

from ..specter_error import SpecterError, handle_exception
from ..util.requests_tools import failsafe_request_get

logger = logging.getLogger(__name__)

OZ_TO_G = 28.3495231

currency_mapping = {
    "usd": {
        "symbol": "$",
        "support": [
            "bitstamp",
            "coindesk",
            "spotbit_coinbase",
            "spotbit_kraken",
            "spotbit_bitfinex",
            "spotbit_okcoin",
            "spotbit_bitstamp",
        ],
    },
    "eur": {
        "symbol": "€",
        "support": [
            "bitstamp",
            "coindesk",
            "spotbit_coinbase",
            "spotbit_kraken",
            "spotbit_bitfinex",
            "spotbit_okcoin_eur",
            "spotbit_bitstamp",
        ],
    },
    "gbp": {
        "symbol": "£",
        "support": [
            "bitstamp",
            "coindesk",
            "spotbit_coinbase",
            "spotbit_kraken",
            "spotbit_bitfinex",
            "spotbit_bitstamp",
        ],
    },
    "chf": {
        "symbol": " Fr.",
        "support": ["coindesk", "spotbit_coinbase", "spotbit_kraken"],
    },
    "aud": {
        "symbol": "$",
        "support": ["coindesk", "spotbit_coinbase", "spotbit_kraken"],
    },
    "cad": {
        "symbol": "$",
        "support": ["coindesk", "spotbit_coinbase", "spotbit_kraken"],
    },
    "nzd": {"symbol": "$", "support": ["coindesk", "spotbit_coinbase"]},
    "hkd": {"symbol": "$", "support": ["coindesk", "spotbit_coinbase"]},
    "jpy": {
        "symbol": "¥",
        "support": [
            "coindesk",
            "spotbit_coinbase",
            "spotbit_kraken",
            "spotbit_bitfinex",
        ],
    },
    "rub": {"symbol": "₽", "support": ["coindesk", "spotbit_coinbase"]},
    "ils": {"symbol": "₪", "support": ["coindesk", "spotbit_coinbase"]},
    "jod": {"symbol": "د.ا", "support": ["coindesk", "spotbit_coinbase"]},
    "twd": {"symbol": "$", "support": ["coindesk", "spotbit_coinbase"]},
    "brl": {"symbol": " BRL", "support": ["coindesk", "spotbit_coinbase"]},
    "xau": {
        "symbol": " oz. ",
        "support": ["coindesk", "spotbit_coinbase"],
        "weight_unit_convertible": True,
    },
    "xag": {
        "symbol": " oz. ",
        "support": ["coindesk", "spotbit_coinbase"],
        "weight_unit_convertible": True,
    },
    "xpt": {
        "symbol": " oz. ",
        "support": ["spotbit_coinbase"],
        "weight_unit_convertible": True,
    },
    "xpd": {
        "symbol": " oz. ",
        "support": ["spotbit_coinbase"],
        "weight_unit_convertible": True,
    },
}


def update_price(specter, current_user):
    try:
        price, symbol = get_price_at(specter, timestamp="now")
        specter.update_alt_rate(price, current_user)
        specter.update_alt_symbol(symbol, current_user)
        return True
    except SpecterError as e:
        logger.error(f"{e} while updating price")
    except Exception as e:
        handle_exception(e)
        return False


"""
    Tries to get the current BTC price based on the user provider preferences.
    Returns: (success, price, symbol)
"""


# (provider, currency) = specter.price_provider.split()


def get_price_at(specter, timestamp="now"):
    """returns a price and a currency-symbol at the timestamp"""
    try:
        if specter.price_check:
            requests_session = specter.requests_session(
                force_tor=("spotbit" in specter.price_provider)
            )
            # something like "spotbit_bitstamp":
            (exchange, currency) = _parse_exchange_currency(specter.price_provider)
            try:
                currency_symbol = currency_mapping[currency]["symbol"]
                weight_unit_convertible = currency_mapping[currency].get(
                    "weight_unit_convertible", False
                )
            except AttributeError:
                raise SpecterError(f"Currency not supported: {currency}")

            if exchange not in currency_mapping[currency]["support"]:
                raise SpecterError(
                    f"The currency {currency} is not supported on exchange {exchange}"
                )

            if specter.price_provider.startswith("bitstamp"):
                if timestamp == "now":
                    price = failsafe_request_get(
                        requests_session,
                        "https://www.bitstamp.net/api/v2/ticker/btc{}".format(currency),
                    )["last"]
                else:
                    price = requests_session.get(
                        "https://www.bitstamp.net/api/v2/ohlc/btc{}/?limit=1&step=86400&start={}".format(
                            currency, timestamp
                        )
                    ).json()["data"]["ohlc"][0]["close"]
            elif specter.price_provider.startswith("coindesk"):
                if timestamp == "now":
                    price = failsafe_request_get(
                        requests_session,
                        f"https://api.coindesk.com/v1/bpi/currentprice/{currency.upper()}.json",
                    )["bpi"][currency.upper()]["rate_float"]
                else:
                    raise SpecterError("coindesk does not support historic prices")
            elif specter.price_provider.startswith("spotbit"):
                exchange = specter.price_provider.split("spotbit_")[1].split("_")[0]
                if timestamp == "now":
                    price = failsafe_request_get(
                        requests_session,
                        "http://h6zwwkcivy2hjys6xpinlnz2f74dsmvltzsd4xb42vinhlcaoe7fdeqd.onion/now/{}/{}".format(
                            currency, exchange
                        ),
                    )["close"]
                else:
                    price = failsafe_request_get(
                        requests_session,
                        "http://h6zwwkcivy2hjys6xpinlnz2f74dsmvltzsd4xb42vinhlcaoe7fdeqd.onion/hist/{}/{}/{}/{}".format(
                            currency,
                            exchange,
                            timestamp * 1000,
                            (timestamp + 121) * 1000,
                        ),
                    )["data"][0][7]
            if weight_unit_convertible:
                if specter.weight_unit == "gram":
                    price = price * OZ_TO_G
                    currency_symbol = " g."
                elif specter.weight_unit == "kg":
                    price = price * OZ_TO_G / 1000
                    currency_symbol = " kg"
            return (price, currency_symbol)
        raise Exception("get_price_at called whereas specter.price_check is False")
    except SpecterError as se:
        raise se
    except KeyError as ke:
        raise SpecterError(f"Error as json doesn't look reasonable: {ke}")


def _parse_exchange_currency(exchange_currency):
    # e.g. "spotbit_bitstamp_eur" or "bitstamp_eur"
    arr = exchange_currency.split("_")
    if len(arr) == 2:
        return arr[0], arr[1]
    elif len(arr) == 3:
        return f"{arr[0]}_{arr[1]}", arr[2]
    raise SpecterError(f"Cannot parse exchange_currency: {exchange_currency}")



from ..service import Service, devstatus_alpha


class BitcoinReserveService(Service):
    id = "bitcoinreserve"
    name = "Bitcoin Reserve"
    icon = "bitcoinreserve/img/ghost.png"  # Need to clarify logo usage "bitcoinreserve/img/bitcoinreserve_icon.svg"
    logo = "bitcoinreserve/img/ghost.png"  # Need to clarify logo usage "bitcoinreserve/img/bitcoinreserve_icon.svg"
    desc = "Where Europe buys Bitcoin."
    has_blueprint = True
    devstatus = devstatus_alpha

    # TODO: As more Services are integrated, we'll want more robust categorization and sorting logic
    sort_priority = 2



class DeviceTypes:
    BITBOX02 = "bitbox02"
    BITCOINCORE = "bitcoincore"
    BITCOINCORE_WATCHONLY = "bitcoincore_watchonly"
    COBO = "cobo"
    COLDCARD = "coldcard"
    ELECTRUM = "electrum"
    ELEMENTSCORE = "elementscore"
    GENERICDEVICE = "other"
    JADE = "jade"
    KEEPKEY = "keepkey"
    LEDGER = "ledger"
    SEEDSIGNER = "seedsigner"
    SPECTERDIY = "specter"
    TREZOR = "trezor"
    KEYSTONE = "keystone"
    PASSPORT = "passport"



import os
import json
import logging
import pathlib
from flask_babel import lazy_gettext as _

from cryptoadvance.specter.devices.bitcoin_core import BitcoinCoreWatchOnly

from ..helpers import alias, load_jsons
from ..rpc import get_default_datadir

from ..devices import __all__ as device_classes
from ..devices.generic import GenericDevice  # default device type
from ..persistence import write_device, delete_file, delete_folder

logger = logging.getLogger(__name__)


def get_device_class(device_type):
    """Look up device class by its type"""
    for cls in device_classes:
        if device_type == cls.device_type:
            return cls
    return GenericDevice


class DeviceManager:
    """A DeviceManager mainly manages the persistence of a device-json-structures
    compliant to helper.load_jsons
    """

    # of them via json-files in an empty data folder
    def __init__(self, data_folder):
        self.update(data_folder=data_folder)

    def update(self, data_folder=None):
        if data_folder is not None:
            self.data_folder = data_folder
            if data_folder.startswith("~"):
                data_folder = os.path.expanduser(data_folder)
            # creating folders if they don't exist
            pathlib.Path(data_folder).mkdir(parents=True, exist_ok=True)
        devices = {}
        devices_files = load_jsons(self.data_folder, key="name")
        for device_alias in devices_files:
            fullpath = os.path.join(self.data_folder, "%s.json" % device_alias)
            devices[devices_files[device_alias]["name"]] = get_device_class(
                devices_files[device_alias]["type"]
            ).from_json(
                devices_files[device_alias],
                self,
                default_alias=device_alias,
                default_fullpath=fullpath,
            )
        self.devices = devices

    @property
    def devices_names(self):
        return sorted(self.devices.keys())

    def add_device(self, name, device_type, keys):
        device_alias = alias(name)
        fullpath = os.path.join(self.data_folder, "%s.json" % device_alias)
        i = 2
        while os.path.isfile(fullpath):
            device_alias = alias("%s %d" % (name, i))
            fullpath = os.path.join(self.data_folder, "%s.json" % device_alias)
            i += 1
        # remove duplicated keys if any exist
        non_dup_keys = []
        for key in keys:
            if key not in non_dup_keys:
                non_dup_keys.append(key)
        keys = non_dup_keys
        device = get_device_class(device_type)(
            name, device_alias, keys, "", fullpath, self
        )
        write_device(device, fullpath)
        self.update()  # reload files
        return device

    def get_by_alias(self, device_alias):
        for device_name in self.devices:
            if self.devices[device_name].alias == device_alias:
                return self.devices[device_name]
        logger.error(_("Could not find Device {}").format(device_alias))

    def remove_device(
        self,
        device,
        wallet_manager=None,
        bitcoin_datadir=get_default_datadir(),
        chain="main",
    ):
        delete_file(device.fullpath)
        # if device can delete itself - call it
        if hasattr(device, "delete"):
            device.delete(wallet_manager, bitcoin_datadir=bitcoin_datadir, chain=chain)
        self.update()

    @property
    def supported_devices(self):
        return device_classes

    def supported_devices_for_chain(self, specter):
        """Devices which you can create via the UI. BitcoinCoreWatchonly is not among them
        and if we have no node, hot-wallets neither. Liquid support is limited as well.
        """
        if not specter.chain:
            devices = [
                device_class
                for device_class in device_classes
                if device_class.device_type not in ["bitcoincore", "elementscore"]
            ]
        elif specter.is_liquid:
            devices = [
                device_class
                for device_class in device_classes
                if device_class.liquid_support
            ]
        else:
            devices = [
                device_class
                for device_class in device_classes
                if device_class.bitcoin_core_support
            ]
        if BitcoinCoreWatchOnly in devices:
            devices.remove(BitcoinCoreWatchOnly)
        return devices

    def delete(self, specter):
        """Deletes all the devices"""
        for d in self.devices:
            device = self.devices[d]
            self.remove_device(device)
        delete_folder(self.data_folder)



from . import DeviceTypes
from ..device import Device


class GenericDevice(Device):
    device_type = DeviceTypes.GENERICDEVICE
    name = "Other"

    sd_card_support = True
    qr_code_support = True
    liquid_support = True
    taproot_support = True

    def create_psbts(self, base64_psbt, wallet):
        # in QR codes keep only xpubs
        qr_psbt = wallet.fill_psbt(base64_psbt, non_witness=False, xpubs=True)
        # in SD card put as much as possible
        sd_psbt = wallet.fill_psbt(base64_psbt, non_witness=True, xpubs=True)
        psbts = {"qrcode": qr_psbt, "sdcard": sd_psbt}
        return psbts



import logging
from functools import wraps

from flask import Blueprint
from flask import current_app as app
from flask import flash, redirect, render_template, request, url_for
from flask_babel import lazy_gettext as _
from flask_login import current_user, login_required

from ..services import ExtensionException

logger = logging.getLogger(__name__)


# This endpoint is just there to share templates between services.
services_endpoint = Blueprint(
    "services_endpoint", __name__, template_folder="templates"
)

# All blueprint from Services are no longer loaded statically but dynamically when the service-class in initialized
# check cryptoadvance.specter.services.service_manager.Service for doing that and
# check cryptoadvance.specter.services/**/manifest for instances of Service-classes and
# check cryptoadvance.specter.services.service_manager.ServiceManager.services for initialisation of ServiceClasses


def user_secret_decrypted_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if app.config["LOGIN_DISABLED"]:
            # No logins means no password so no user_secret is possible
            flash(
                _(
                    "Service integration requires an authentication method that includes a password"
                )
            )
            return redirect(url_for(f"settings_endpoint.auth"))
        elif not current_user.is_user_secret_decrypted:
            flash(_("Must login again to enable protected Services-related data"))
            # Force re-login; automatically redirects back to calling page
            return app.login_manager.unauthorized()
        else:
            return func(*args, **kwargs)

    return wrapper


@services_endpoint.route("/choose", methods=["GET"])
def choose():
    return render_template(
        "services/choose.jinja",
        is_login_disabled=app.config["LOGIN_DISABLED"],
        specter=app.specter,
        services=app.specter.service_manager.services_sorted,
    )


@services_endpoint.route(
    "/associate_addr/<wallet_alias>/<address>", methods=["GET", "POST"]
)
@login_required
@user_secret_decrypted_required
def associate_addr(wallet_alias, address):
    wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)

    if request.method == "POST":
        service_id = request.form["service_id"]
        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
        service_cls = app.specter.service_manager.get_service(service_id)
        service_cls.reserve_address(wallet=wallet, address=address)
        return redirect(
            url_for("wallets_endpoint.addresses", wallet_alias=wallet_alias)
        )

    addr_obj = wallet.get_address_obj(address=address)

    # Inject the User's active Services
    services = []
    for service_id in current_user.services:
        try:
            services.append(app.specter.service_manager.get_service(service_id))
        except ExtensionException:
            pass

    return render_template(
        "services/associate_addr.jinja",
        specter=app.specter,
        services=services,
        wallet=wallet,
        addr_obj=addr_obj,
    )



from datetime import datetime
from flask import current_app as app
from flask import Blueprint
from jinja2 import pass_context
from ..helpers import to_ascii20
from ..util.common import format_btc_amount_as_sats, format_btc_amount

filters_bp = Blueprint("filters", __name__)

############### filters ##################


@pass_context
@filters_bp.app_template_filter("ascii20")
def ascii20(context, name):
    return to_ascii20(name)


@pass_context
@filters_bp.app_template_filter("unique_len")
def unique_len(context, arr):
    return len(set(arr))


@pass_context
@filters_bp.app_template_filter("datetime")
def timedatetime(context, s):
    return format(datetime.fromtimestamp(s), "%d.%m.%Y %H:%M")


@pass_context
@filters_bp.app_template_filter("average_of_attribute")
def average_of_attribute(context, values, attribute):
    dicts = [
        getattr(value, attribute)
        for value in values
        if getattr(value, attribute) is not None
    ]
    return sum(dicts) / len(dicts) if dicts else None


@pass_context
@filters_bp.app_template_filter("btcunitamount_fixed_decimals")
def btcunitamount_fixed_decimals(
    context,
    value,
    maximum_digits_to_strip=7,
    minimum_digits_to_strip=6,
    enable_digit_formatting=True,
):
    if app.specter.hide_sensitive_info:
        return "#########"
    if value is None:
        return "Unknown"
    if value < 0 and app.specter.is_liquid:
        return "Confidential"
    if app.specter.unit == "sat":
        return format_btc_amount_as_sats(value)

    return format_btc_amount(
        value,
        maximum_digits_to_strip=maximum_digits_to_strip,
        minimum_digits_to_strip=minimum_digits_to_strip,
        enable_digit_formatting=enable_digit_formatting,
    )


@pass_context
@filters_bp.app_template_filter("btcamount")
def btcamount(context, value):
    if value is None:
        return "Unknown"
    if value < 0 and app.specter.is_liquid:
        return "Confidential"
    value = round(float(value), 8)
    return "{:,.8f}".format(value).rstrip("0").rstrip(".")


@pass_context
@filters_bp.app_template_filter("btc2sat")
def btc2sat(context, value):
    value = int(round(float(value) * 1e8))
    return f"{value}"


@pass_context
@filters_bp.app_template_filter("feerate")
def feerate(context, value):
    value = float(value) * 1e8
    # workaround for minimal fee rate
    # because 1.01 doesn't look nice
    if value <= 1.02:
        value = 1
    return "{:,.2f}".format(value).rstrip("0").rstrip(".")


@pass_context
@filters_bp.app_template_filter("btcunitamount")
def btcunitamount(context, value):
    if app.specter.hide_sensitive_info:
        return "#########"
    if value is None:
        return "Unknown"
    if value < 0 and app.specter.is_liquid:
        return "Confidential"
    if app.specter.unit != "sat":
        return btcamount(context, value)
    value = float(value)
    return "{:,.0f}".format(round(value * 1e8))


@pass_context
@filters_bp.app_template_filter("altunit")
def altunit(context, value):
    if app.specter.hide_sensitive_info:
        return "########"
    if value is None:
        return "Can't be calculated"
    if value < 0:
        return "-"
    if app.specter.price_check and (app.specter.alt_rate and app.specter.alt_symbol):
        rate = (
            "{:,.2f}".format(float(value) * float(app.specter.alt_rate))
            .rstrip("0")
            .rstrip(".")
        )
        if app.specter.alt_symbol in ["$", "£"]:
            return app.specter.alt_symbol + rate
        else:
            return rate + app.specter.alt_symbol
    return ""


@pass_context
@filters_bp.app_template_filter("bytessize")
def bytessize(context, value):
    value = float(value)
    return "{:,.0f}".format(value / float(1 << 30)) + " GB"


@pass_context
@filters_bp.app_template_filter("assetlabel")
def assetlabel(context, asset):
    if app.specter.hide_sensitive_info:
        return "####"
    return app.specter.asset_label(asset)



"""
liveness and readiness probes are a semi-standard way of health-checking.
See e.g. here:
https://kubernetes.io/docs/tasks/configure-pod-container/configure-liveness-readiness-startup-probes/
"""
import json
import logging
from os import abort

from cryptoadvance.specter.api.rest.base import BaseResource, rest_resource
from flask import current_app as app

logger = logging.getLogger(__name__)


@rest_resource
class ResourceLiveness(BaseResource):
    """/api/healthz/liveness
    Whether the app is up and running although it might not have connection to DB/nodes etc.
    """

    endpoints = ["/healthz/liveness"]

    def get(self):
        return {"message": "i am alive"}


@rest_resource
class ResourceReadyness(BaseResource):
    """/api/healthz/readyness
    Whether the app is up and running AND ALL its dependent services (in our case nodes) are properly functioning as well.
    """

    endpoints = ["/healthz/readyness"]

    def get(self):
        try:
            # Not sure whether that's enough. Probably improvable:
            app.specter.check()
        except Exception as e:
            abort(500, message="Readyness probe failed")
        return {"message": "i am ready"}



""" util stuff for searching the filesystem mainly used by the reflection.py """

import logging
import os
import sys
from pathlib import Path
from typing import List

logger = logging.getLogger(__name__)


def detect_extension_style_in_cwd(cwd=".") -> str:
    """don't override the cwd other than in testing!!
    if you do, you probably have to mess with sys.path at the same time
    """
    if Path(cwd, "src/cryptoadvance/specter").is_dir() or getattr(sys, "frozen", False):
        return "specter-desktop"
    if Path(cwd, "src").is_dir():
        if Path(cwd, "setup.py").is_file():
            return "publish-ready"
        else:
            raise Exception(
                f"""
                You have an inconsistent project file-layout in folder
                {Path(cwd).resolve()}
                Either you:
                * Have a src-folder and you can have .py-files in your projectroot OR
                * you don't have a src-folder but ./extensionid/service.py (+ __init__.py)
                But not having a ./src-folder AND some .py-file in the project-root is not allowed.
                """
            )
    else:
        # Ad-Hoc style is: ./extensionid/service.py but no .py-files in cwd
        # as this screws up the discovery
        for pth in Path(cwd).iterdir():
            if pth.suffix == ".py":
                raise Exception(
                    f"""
                You have an inconsistent project file-layout in folder
                {Path(cwd).resolve()}
                Either you:
                * Have a src-folder and you can have .py-files in your projectroot OR
                * you don't have a src-folder but ./extensionid/service.py (+ __init__.py)
                But not having a ./src-folder AND some .py-file in the project-root is not allowed.
                """
                )
        return "adhoc"


def search_dirs_in_path(
    path: Path, search_dirname="specterext", return_without_extid=True
) -> List[Path]:
    """recursively walks the filesystem collecting directories which are called "specterext"
    returns a list of PATH all ending with specterext.
    If return_without_extid is False, it'll return one level deeper which might be a different length
    """
    if isinstance(path, str):
        path = Path(path)
    plist: List[Path] = []
    if not path.is_dir():
        raise Exception(f"Search path does not exist: {path}")
    for root, dirs, _ in os.walk(path, followlinks=True):
        for dirname in dirs:
            if dirname == search_dirname:
                if return_without_extid:
                    plist.append(Path(root, dirname))
                else:
                    subdirs = [x for x in Path(root, dirname).iterdir() if x.is_dir()]
                    plist.extend(subdirs)
    return plist



import base64
import binascii
import cryptography
import hashlib
import json
import logging
import os
import shutil

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from flask_login import UserMixin

from .specter_error import SpecterError, handle_exception
from .persistence import read_json_file, write_json_file, delete_folder
from .managers.wallet_manager import WalletManager
from .managers.device_manager import DeviceManager
from .helpers import deep_update


logger = logging.getLogger(__name__)


def hash_password(plaintext_password):
    """Hash a password for storing."""
    salt = binascii.b2a_base64(hashlib.sha256(os.urandom(60)).digest()).strip()
    pwdhash = (
        binascii.b2a_base64(
            hashlib.pbkdf2_hmac(
                "sha256", plaintext_password.encode("utf-8"), salt, 10000
            )
        )
        .strip()
        .decode()
    )
    return {"salt": salt.decode(), "pwdhash": pwdhash}


def verify_password(stored_password, provided_password):
    """Verify a stored password against one provided by user"""
    pwdhash = hashlib.pbkdf2_hmac(
        "sha256",
        provided_password.encode("utf-8"),
        stored_password["salt"].encode(),
        10000,
    )
    return pwdhash == binascii.a2b_base64(stored_password["pwdhash"])


class UserSecretException(Exception):
    pass


class User(UserMixin):
    """
    The user_secret is used to encrypt/decrypt other user-specific data
    (e.g. services data). It is encrypted for storage using the user's
    password (the method == "none" auth type cannot enable options that
    require encrypted storage because it has no password). user_secret is
    decrypted on login (see: SpecterFlask.login()) and stored in memory
    in plaintext.
    """

    def __init__(
        self,
        id,
        username,
        password_hash,
        config,
        specter,
        encrypted_user_secret=None,
        is_admin=False,
        services=[],
    ):
        self.id = id
        self.username = username
        self.password_hash = password_hash
        self.config = config
        self.encrypted_user_secret = encrypted_user_secret
        self.plaintext_user_secret = None
        self.is_admin = is_admin
        self.uid = specter.config["uid"]
        self.specter = specter
        self.wallet_manager = None
        self.device_manager = None
        self.manager = None
        self._services = services

        # Iterations will need to be increased over time to keep ahead of CPU advances.
        self.encryption_iterations = 390000

    # TODO: User obj instantiation belongs in UserManager
    @classmethod
    def from_json(cls, user_dict, specter):
        try:
            user_args = {
                "id": user_dict["id"],
                "username": user_dict["username"],
                "password_hash": user_dict[
                    "password"
                ],  # TODO: Migrate attr name to "password_hash"?
                "config": {},
                "specter": specter,
                "encrypted_user_secret": user_dict.get("encrypted_user_secret", None),
                "services": user_dict.get("services", []),
            }
            if not user_dict["is_admin"]:
                user_args["config"] = user_dict["config"]
                return cls(**user_args)
            else:
                user_args["is_admin"] = True
                return cls(**user_args)

        except Exception as e:
            handle_exception(e)
            raise SpecterError(f"Unable to parse user JSON.:{e}")

    @property
    def folder_id(self):
        if self.is_admin:
            return ""
        return f"_{self.id}"

    @property
    def services(self):
        if not self._services:
            self._services = []
        return self._services

    @property
    def is_user_secret_decrypted(self):
        return self.plaintext_user_secret is not None

    def _encrypt_user_secret(self, plaintext_password):
        """
        Implementation taken from the pyca/cryptography docs:
        https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet
        """
        salt = os.urandom(16)

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=self.encryption_iterations,
        )
        key = base64.urlsafe_b64encode(kdf.derive(plaintext_password.encode()))
        f = Fernet(key)
        token = f.encrypt(self.plaintext_user_secret)

        self.encrypted_user_secret = {
            "token": token.decode(),
            "salt": base64.b64encode(salt).decode(),
            "iterations": self.encryption_iterations,
        }

    def decrypt_user_secret(self, plaintext_password):
        # see: https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet
        if not self.encrypted_user_secret:
            self._generate_user_secret(plaintext_password)

        token = self.encrypted_user_secret["token"].encode()
        salt = base64.b64decode(self.encrypted_user_secret["salt"])
        iterations = self.encrypted_user_secret["iterations"]

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=iterations,
        )
        key = base64.urlsafe_b64encode(kdf.derive(plaintext_password.encode()))
        f = Fernet(key)

        self.plaintext_user_secret = f.decrypt(token)

        # If this encrypted_user_secret has an outdated (weaker) number of iterations,
        #   re-encrypt with the current higher iteration count.
        if iterations < self.encryption_iterations:
            self._encrypt_user_secret(plaintext_password)

    def _generate_user_secret(self, plaintext_password):
        """
        Generates and stores the user_secret in memory. Also stores it encrypted
        to disk.

        Encryption using the user_secret uses a Fernet key. But the Fernet
        key itself will be encrypted with the user's password.
        """
        self.plaintext_user_secret = Fernet.generate_key()
        self._encrypt_user_secret(plaintext_password)
        self.save_info()
        logger.debug("Generated user_secret")

    def delete_user_secret(self, autosave: bool = True):
        self.encrypted_user_secret = None
        self.plaintext_user_secret = None
        if autosave:
            self.save_info()

    def set_password(self, plaintext_password):
        """Hash the incoming plaintext password and update the encrypted user_secret as
        needed.

        Remember that the underlying user_secret doesn't change if the user changes
        their password; it's the same user_secret but it just needs to be
        re-encrypted using the new password.
        """
        # Check the encrypted_user_secret before saving password change!
        if self.encrypted_user_secret is None:
            # First time this user is initializing their user_secret
            self._generate_user_secret(plaintext_password)
        else:
            if self.plaintext_user_secret is None:
                # encrypted_user_secret hasn't been decrypted in memory; try to decrypt
                # it now (will only work if we're re-enabling the same password for the
                # admin account).
                try:
                    self.decrypt_user_secret(plaintext_password)
                except cryptography.fernet.InvalidToken as e:
                    # Existing encrypted_user_secret cannot be decrypted with this new
                    # password! Alert the calling code to handle (either provide the
                    # previous password to decrypt/re-encrypt or delete all existing
                    # encrypted data for this user (admin) because it's no longer
                    # decryptable).
                    logger.warn(e)
                    raise UserSecretException(
                        "Cannot decrypt existing encrypted_user_secret with the provided password"
                    )
            else:
                # Must keep re-encrypt encrypted_user_secret with the new password.
                self._encrypt_user_secret(plaintext_password)

        self.password_hash = hash_password(plaintext_password)

    @property
    def json(self):
        user_dict = {
            "id": self.id,
            "username": self.username,
            "password": self.password_hash,  # TODO: Migrate attr name to "password_hash"?
            "is_admin": self.is_admin,
            "encrypted_user_secret": self.encrypted_user_secret,
            "services": self.services,
        }
        if not self.is_admin:
            user_dict["config"] = self.config
        return user_dict

    def check(self):
        self.check_device_manager()
        self.check_wallet_manager()

    def check_wallet_manager(self):
        """Updates wallet manager for this user"""
        # if chain, user or data folder changed
        wallet_manager = self.wallet_manager
        wallets_rpcpath = "specter%s" % self.uid
        wallets_folder = os.path.join(
            self.specter.data_folder, f"wallets{self.folder_id}"
        )
        if (
            wallet_manager is None
            or wallet_manager.data_folder != wallets_folder
            or wallet_manager.rpc_path != wallets_rpcpath
            or wallet_manager.chain != self.specter.chain
        ):

            wallet_manager = WalletManager(
                self.specter.bitcoin_core_version_raw,
                wallets_folder,
                self.specter.rpc,
                self.specter.chain,
                self.device_manager,
                path=wallets_rpcpath,
            )
            self.wallet_manager = wallet_manager
        else:
            wallet_manager.update(
                wallets_folder, self.specter.rpc, chain=self.specter.chain
            )

    def check_device_manager(self, user=None):
        """Updates device manager for this user"""
        devices_folder = os.path.join(
            self.specter.data_folder, f"devices{self.folder_id}"
        )
        if self.device_manager is None:
            self.device_manager = DeviceManager(devices_folder)
        else:
            self.device_manager.update(data_folder=devices_folder)

    # TODO: Refactor this into UserManager
    def save_info(self, delete=False):
        if self.manager is None:
            self.manager = self.specter.user_manager
        users = self.manager.users
        existing = self in users

        # update specter users
        if not existing and not delete:
            self.specter.user_manager.add_user(self)
        if existing and delete:
            self.specter.delete_user(self)
        self.manager.save()

    def add_service(self, service_id: str, autosave: bool = True):
        """Add a Service to the User. Only updates what is listed in the sidebar."""
        if service_id not in self._services:
            self._services.append(service_id)
        if autosave:
            self.save_info()

    def has_service(self, service_id: str) -> bool:
        """Returns true if the User has that service"""
        return service_id in self._services

    def remove_service(self, service_id: str, autosave: bool = True):
        """Remove a Service from the User. Only updates what is listed in the sidebar."""
        if service_id in self.services:
            self.services.remove(service_id)
        if autosave:
            self.save_info()

    # TODO: Refactor calling code to explicitly call User.save() rather than embedding
    #   self.save_info() on every update and setter. It ends up saving to disk multiple
    #   times for a single Settings submit.
    def update_asset_label(self, asset, label, chain):
        if "asset_labels" not in self.config:
            self.config["asset_labels"] = {}
        if not label:
            if self.config["asset_labels"].get(chain, {}).get(asset):
                del self.config["asset_labels"][chain][asset]
        else:
            deep_update(self.config["asset_labels"], {chain: {asset: label}})
        self.save_info()

    def set_explorer(self, explorer_id, explorer):
        if "explorers" not in self.config:
            self.config["explorers"] = (
                {"main": "", "test": "", "regtest": "", "signet": ""},
            )
        self.config["explorers"][self.specter.chain] = explorer
        if "explorer_id" not in self.config:
            self.config["explorer_id"] = {
                "main": "CUSTOM",
                "test": "CUSTOM",
                "regtest": "CUSTOM",
                "signet": "CUSTOM",
            }
        self.config["explorer_id"][self.specter.chain] = explorer_id
        self.save_info()

    def set_fee_estimator(self, fee_estimator, custom_url):
        self.config["fee_estimator"] = fee_estimator
        if fee_estimator == "custom":
            self.config["fee_estimator_custom_url"] = custom_url
        self.save_info()

    def set_hwi_bridge_url(self, url):
        self.config["hwi_bridge_url"] = url
        self.save_info()

    def set_unit(self, unit):
        self.config["unit"] = unit
        self.save_info()

    def set_price_check(self, price_check_bool):
        self.config["price_check"] = price_check_bool
        self.save_info()

    def set_hide_sensitive_info(self, hide_sensitive_info_bool):
        self.config["hide_sensitive_info"] = hide_sensitive_info_bool
        self.save_info()

    def set_autohide_sensitive_info_timeout(self, timeout_minutes):
        self.config["autohide_sensitive_info_timeout_minutes"] = timeout_minutes
        self.save_info()

    def set_autologout_timeout(self, timeout_hours):
        self.config["autologout_timeout_hours"] = timeout_hours
        self.save_info()

    def set_price_provider(self, price_provider):
        self.config["price_provider"] = price_provider
        self.save_info()

    def set_weight_unit(self, weight_unit):
        self.config["weight_unit"] = weight_unit
        self.save_info()

    def set_alt_rate(self, alt_rate):
        self.config["alt_rate"] = alt_rate
        self.save_info()

    def set_alt_symbol(self, alt_symbol):
        self.config["alt_symbol"] = alt_symbol
        self.save_info()

    def delete(self):
        # we delete wallet manager and device manager in save_info
        self.save_info(delete=True)

    def __eq__(self, other):
        if other == None:
            return False
        if isinstance(other, str):
            return self.id == other
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # to make lookups in dicts by user id
        return hash(self.id)

    def __str__(self):
        return self.id

    def __repr__(self):
        return f"User({self.id})"



import json
import logging
import os

from cryptoadvance.specter.managers.genericdata_manager import GenericDataManager
from cryptoadvance.specter.managers.singleton import ConfigurableSingleton
from cryptoadvance.specter.specter_error import SpecterError
from cryptoadvance.specter.user import User

logger = logging.getLogger("__name__")


class ServiceEncryptedStorageError(Exception):
    pass


class ServiceEncryptedStorage(GenericDataManager):
    """
    Encrypted storage class for *ALL* Services-related secrets and config for a given user.
    The json file is stored as: <username>_services.json

    Each Service may specify its own json data format.

    Data is stored internally as a json string for easier encryption handling to disk.

    Note that this is storage is separate from the un-encrypted ServiceAnnotationsStorage.
    This storage is meant for secrets that must be encrypted and any Service-related config
    that the User might need to store (e.g. which Wallet is attached to the Service). This
    config data doesn't need to be encrypted, but we may as well store it here rather than
    adding data bloat to the User json.

    * `disable_decrypt`: Allows the ServiceEncryptedStorageManager to see which Services
        have data but without attempting to decrypt the values. This allows us to check
        if there is encrypted Service data even when we don't have the
        plaintext_user_secret.
    """

    def __init__(self, data_folder: str, user: User, disable_decrypt: bool = False):

        if not user.plaintext_user_secret and not disable_decrypt:
            raise ServiceEncryptedStorageError(
                f"User {user} must be authenticated with password before encrypted service data can be loaded"
            )

        # Must set the user before calling the parent's __init__(); it calls load()
        #   which then calls data_file below.
        self.user = user

        if disable_decrypt:
            super().__init__(data_folder, encryption_key=None)
        else:
            super().__init__(data_folder, encryption_key=user.plaintext_user_secret)

    @property
    def encrypted_fields(self):
        """We override the class member to force ALL data fields to be considered
        encrypted."""
        fields = list(self.data.keys())
        if "encrypted_storage_version" in fields:
            fields.pop(fields.index("encrypted_storage_version"))
        return fields

    @property
    def data_file(self):
        return os.path.join(self.data_folder, f"{self.user.username}_services.json")

    def set_service_data(self, service_id: str, data: dict, autosave: bool = True):
        """Store the api_data json blob as a string; completely overwrites previous state"""
        if data == {} and service_id in self.data:
            del self.data[service_id]
        else:
            self.data[service_id] = json.dumps(data)
        if autosave:
            self._save()

    def update_service_data(self, service_id: str, data: dict, autosave: bool = True):
        if data == {}:
            logger.debug("This is a nonsense no-op")
            return

        # Add or update fields; does not remove existing fields
        if service_id not in self.data:
            # Initialize a blank entry
            cur_data = {}
        else:
            cur_data = json.loads(self.data[service_id])
        cur_data.update(data)
        self.data[service_id] = json.dumps(cur_data)
        if autosave:
            self._save()

    def get_service_data(self, service_id: str) -> dict:
        service_data = self.data.get(service_id, None)
        if service_data:
            # Convert string back into json blob
            service_data = json.loads(service_data)
        else:
            service_data = {}
        return service_data


class ServiceUnencryptedStorage(ServiceEncryptedStorage):
    """In order to use ServiceEncryptedStorage but unencrypted, we derive from that class
    and change the datafile.
    """

    def __init__(self, data_folder: str, user: User, disable_decrypt: bool = False):
        if not disable_decrypt:
            raise Exception(
                "ServiceUnencryptedStorage needs to be initialized with disable_decrypt = True"
            )
        if disable_decrypt:
            super().__init__(data_folder, encryption_key=None, disable_decrypt=True)

    @property
    def data_file(self):
        return os.path.join(
            self.data_folder, f"{self.user.username}_unencrypted_services.json"
        )


class ServiceEncryptedStorageManager(ConfigurableSingleton):
    """Singleton that manages access to users' ServiceApiKeyStorage; context-aware so it
    knows who the current_user is for the given request context.

    Requires a one-time configuration call on startup in the ServiceManager.
    """

    @classmethod
    def configure_instance(cls, data_folder, user_manager):
        super().configure_instance()
        cls._instance.data_folder = data_folder
        cls._instance.user_manager = user_manager
        cls._instance.storage_by_user = {}

    def get_raw_encrypted_data(self, user: User) -> dict:
        """Doesn't attempt to decrypt the ServiceEncryptedStorage, just returns the
        user's full encrypted Service data json as-is."""
        return ServiceEncryptedStorage(
            self.data_folder, user, disable_decrypt=True
        ).data

    def _get_current_user_service_storage(self) -> ServiceEncryptedStorage:
        """Returns the storage-class for the current_user. Lazy_init if necessary"""
        user = self.user_manager.get_user()

        if user not in self.storage_by_user:
            self.storage_by_user[user] = ServiceEncryptedStorage(self.data_folder, user)
        return self.storage_by_user[user]

    def set_current_user_service_data(self, service_id: str, service_data: dict):
        self._get_current_user_service_storage().set_service_data(
            service_id, service_data
        )

    def update_current_user_service_data(self, service_id: str, service_data: dict):
        # Add or update fields; does not remove existing fields
        self._get_current_user_service_storage().update_service_data(
            service_id, service_data
        )

    def get_current_user_service_data(self, service_id: str) -> dict:
        service_storage = self._get_current_user_service_storage()
        if service_storage:
            return service_storage.get_service_data(service_id)

    def unload_current_user(self):
        """Clear user's ServiceEncryptedStorage from memory (but it remains safely on disk)"""
        user = self.user_manager.get_user()
        if user and user in self.storage_by_user:
            self.storage_by_user[user] = None

    def delete_all_service_data(self, user: User):
        """Completely removes all data in the User's ServiceEncryptedStorage from memory and on-disk."""
        # Clear it from memory...
        self.storage_by_user.pop(user, None)

        # ...and wipe the on-disk storage
        encrypted_storage = ServiceEncryptedStorage(
            self.data_folder, user, disable_decrypt=True
        )
        encrypted_storage.data = {}
        encrypted_storage._save()


class ServiceUnencryptedStorageManager(ServiceEncryptedStorageManager):
    def __init__(self, user_manager, data_folder):
        self.user_manager = user_manager
        self.data_folder = data_folder
        self.storage_by_user = {}

    def _get_current_user_service_storage(self) -> ServiceEncryptedStorage:
        """Returns the storage-class for the current_user. Lazy_init if necessary"""
        user = self.user_manager.get_user()

        if user not in self.storage_by_user:
            self.storage_by_user[user] = ServiceUnencryptedStorage(
                self.data_folder, user, disable_decrypt=True
            )
        return self.storage_by_user[user]



from ..addresslist import *
from embit.liquid.addresses import addr_decode, to_unconfidential


class LAddress(Address):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._unconfidential = to_unconfidential(self.address)

    @property
    def unconfidential(self):
        return self._unconfidential or self.address

    @property
    def is_confidential(self):
        return self.address != self.unconfidential


class LAddressList(AddressList):
    AddressCls = LAddress

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # scriptpubkey dict for lookups of unconf addresses
        self._scripts = {}
        self._update_scripts()

    def _update_scripts(self):
        for addr in list(self.keys()):
            sc, _ = addr_decode(addr)
            if sc and sc not in self._scripts:
                self._scripts[sc] = super().__getitem__(addr)

    def add(self, *args, **kwargs):
        res = super().add(*args, **kwargs)
        # update scriptpubkey dict for lookups of unconf addresses
        self._update_scripts()
        return res

    def __contains__(self, addr):
        """finds address by confidential or unconfidential address by converting to scriptpubkey"""
        try:  # can fail if addr is "Fee", "Dummy" or hex-scriptpubkey
            sc, _ = addr_decode(addr)
            if sc and self._scripts.__contains__(sc):
                return True
        except:
            pass
        return super().__contains__(addr)

    def __getitem__(self, addr):
        """finds address by confidential or unconfidential address by converting to scriptpubkey"""
        sc, _ = addr_decode(addr)
        if sc in self._scripts:
            return self._scripts[sc]
        return super().__getitem__(addr)

    def get(self, addr, default=None):
        try:
            return self[addr]
        except KeyError:
            return default
        except TypeError:
            logger.warning(f"{addr} seems to be invalid")
            return default



import multiprocessing
from cryptoadvance.specter.server import init_app, create_app, create_and_init

from gunicorn.app.wsgiapp import WSGIApplication


class SpecterGunicornApp(WSGIApplication):
    def __init__(
        self, config="cryptoadvance.specter.config.ProductionConfig", options=None
    ):
        """
        i guess those are the potential options:
        usage: __main__.py [-h] [-v] [-c CONFIG] [-b ADDRESS] [--backlog INT] [-w INT] [-k STRING] [--threads INT]
               [--worker-connections INT] [--max-requests INT] [--max-requests-jitter INT] [-t INT] [--graceful-timeout INT]
               [--keep-alive INT] [--limit-request-line INT] [--limit-request-fields INT] [--limit-request-field_size INT]
               [--reload] [--reload-engine STRING] [--reload-extra-file FILES] [--spew] [--check-config] [--print-config]
               [--preload] [--no-sendfile] [--reuse-port] [--chdir CHDIR] [-D] [-e ENV] [-p FILE] [--worker-tmp-dir DIR]
               [-u USER] [-g GROUP] [-m INT] [--initgroups] [--forwarded-allow-ips STRING] [--access-logfile FILE]
               [--disable-redirect-access-to-syslog] [--access-logformat STRING] [--error-logfile FILE] [--log-level LEVEL]
               [--capture-output] [--logger-class STRING] [--log-config FILE] [--log-syslog-to SYSLOG_ADDR] [--log-syslog]
               [--log-syslog-prefix SYSLOG_PREFIX] [--log-syslog-facility SYSLOG_FACILITY] [-R] [--statsd-host STATSD_ADDR]
               [--dogstatsd-tags DOGSTATSD_TAGS] [--statsd-prefix STATSD_PREFIX] [-n STRING] [--pythonpath STRING]
               [--paste STRING] [--proxy-protocol] [--proxy-allow-from PROXY_ALLOW_IPS] [--keyfile FILE] [--certfile FILE]
               [--ssl-version SSL_VERSION] [--cert-reqs CERT_REQS] [--ca-certs FILE] [--suppress-ragged-eofs]
               [--do-handshake-on-connect] [--ciphers CIPHERS] [--paste-global CONF] [--strip-header-spaces]
        """
        self.options = options or {}
        self.config = config
        super().__init__()

    def load(self):
        return create_and_init(self.config)



import os
import time

from ..persistence import read_json_file, write_json_file
from .genericdata_manager import GenericDataManager


class OtpManager(GenericDataManager):
    """
    The OtpManager manages otps (one time passwords) in otps.json
    """

    name_of_json_file = "otps.json"

    def add_new_user_otp(self, otp_dict):
        """adds an OTP for user registration"""
        if "new_user_otps" not in self.data:
            self.data["new_user_otps"] = []
        self.data["new_user_otps"].append(otp_dict)
        self._save()

    def validate_new_user_otp(self, otp):
        """validates an OTP for user registration and removes it if expired"""
        if "new_user_otps" not in self.data:
            return False
        now = time.time()
        for i, otp_dict in enumerate(self.data["new_user_otps"]):
            if otp_dict["otp"] == otp:
                if (
                    "expiry" in otp_dict
                    and otp_dict["expiry"] < now
                    and otp_dict["expiry"] > 0
                ):
                    del self.data["new_user_otps"][i]
                    self._save()
                    return False
                return True
        return False

    def remove_new_user_otp(self, otp):
        """removes an OTP for user registration"""
        if "new_user_otps" not in self.data:
            return False
        for i, otp_dict in enumerate(self.data["new_user_otps"]):
            if otp_dict["otp"] == otp:
                del self.data["new_user_otps"][i]
                self._save()
                return True
        return False



from babel.messages import frontend as babel
from glob import glob

from setuptools import find_namespace_packages, setup

from setuptools.command.install import install


class InstallWithBabelCompile(install):
    """from this stackoverflow question
    https://stackoverflow.com/questions/40051076/compile-translation-files-when-calling-setup-py-install
    """

    def run(self):
        from babel.messages.frontend import compile_catalog

        compiler = compile_catalog(self.distribution)
        option_dict = self.distribution.get_option_dict("compile_catalog")
        compiler.domain = [option_dict["domain"][1]]
        compiler.directory = option_dict["directory"][1]
        compiler.run()
        super().run()


with open("requirements.txt") as f:
    install_reqs = f.read().strip().split("\n")

# Filter out comments/hashes
reqs = []
for req in install_reqs:
    if req.startswith("#") or req.startswith("    --hash="):
        continue
    reqs.append(str(req).rstrip(" \\"))


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="cryptoadvance.specter",
    version="v1.13.0-pre7",
    author="Stepan Snigirev, Kim Neunert",
    author_email="snigirev.stepan@gmail.com, kim.neunert@gmail.com",
    description="A GUI for Bitcoin Core optimised to work with airgapped hardware wallets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cryptoadvance/specter-desktop",
    packages=find_namespace_packages("src", include=["cryptoadvance.*"]),
    package_dir={"": "src"},
    package_data={
        "": [
            "translations/*/LC_MESSAGES/messages.mo",
        ]
    },
    # take METADATA.in into account, include that stuff as well (static/templates)
    include_package_data=True,
    install_requires=reqs,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: Flask",
    ],
    python_requires=">=3.7,<4.0",
    cmdclass={
        "install": InstallWithBabelCompile,
        # The rest is convenience but not strictly necessary for the automation:
        "compile_catalog": babel.compile_catalog,
        "extract_messages": babel.extract_messages,
        "init_catalog": babel.init_catalog,
        "update_catalog": babel.update_catalog,
    },
    entry_points="""
        [console_scripts]
        specter=cryptoadvance.specter.cli:entry_point
    """,
)



import os
import json
import logging
from ..persistence import read_json_file, write_json_file
from ..user import User, hash_password
from flask_login import current_user

logger = logging.getLogger(__name__)


class UserManager:
    """
    The UserManager can manage users
    """

    # of them via json-files in an empty data folder
    def __init__(self, specter):
        self.specter = specter
        self.data_folder = specter.data_folder
        self.update()

    @property
    def users_file(self):
        return os.path.join(self.data_folder, "users.json")

    def load_users(self):
        # if users.json file exists - load from it
        if os.path.isfile(self.users_file):
            users = read_json_file(self.users_file)
        # otherwise - create one and assign unique id
        else:
            users = [
                {
                    "id": "admin",
                    "username": "admin",
                    "password": hash_password("admin"),
                    "is_admin": True,
                    "encrypted_user_secret": None,
                }
            ]
        # convert to User instances
        self.users = [User.from_json(u, self.specter) for u in users]
        if not os.path.isfile(self.users_file):
            self.save()

    def update(self):
        self.load_users()

    def save(self):
        users_json = [u.json for u in self.users]
        write_json_file(users_json, self.users_file)

    def add_user(self, user):
        """Adds a User-Object to the list"""
        if user not in self.users:
            self.users.append(user)
            user.manager = self
        self.save()  # save files
        user.check()
        return self.get_user(user)

    def create_user(self, user_id, username, plaintext_password, config):
        password_hash = hash_password(plaintext_password)
        user = User(user_id, username, password_hash, config, self.specter)
        user.decrypt_user_secret(plaintext_password)
        return self.add_user(user)

    @property
    def admin(self):
        """There is always one admin"""
        for u in self.users:
            if u.is_admin:
                return u

    def get_user(self, user=None):
        """
        Converts from flask_login user to a User in the system.
        Admin by default.
        """
        # get by string
        if user is None:
            user = current_user
        if isinstance(user, str):
            return self.get_by_uid(user)
        if user and not user.is_anonymous:
            if user in self.users:
                return self.get_by_uid(user.id)
        return self.admin

    def get_user_by_username(self, username):
        for u in self.users:
            if u.username == username:
                return u

    @property
    def user(self):
        """User in current context, admin if no context"""
        return self.get_user(current_user)

    def get_by_uid(self, uid):
        for u in self.users:
            if u.id == uid:
                return u
        logger.error("Could not find user %s" % uid)

    def delete_user(self, user):
        self.users.remove(user)
        self.save()



import logging
from mnemonic.mnemonic import ConfigurationError, Mnemonic
from embit import bip32, bip39, networks
from embit.wordlists.bip39 import WORDLIST
from cryptoadvance.specter.specter_error import SpecterError

logger = logging.getLogger(__name__)


""" This module is using two libraries: Trezor's python-mnemonic (https://github.com/trezor/python-mnemonic) 
and embit (https://github.com/diybitcoinhardware/embit)"""


MNEMONIC_LANGUAGES = {
    "en": "english",
    "es": "spanish",
    # "fr": "french", # Not supported, for details see: https://github.com/trezor/python-mnemonic/issues/98)
    "it": "italian",
    # "jp": "japanese",
    # "ko": korean",
    # "?": chinese_simplified",
    # "?": chinese_traditional",
}


def initialize_mnemonic(language_code: str) -> Mnemonic:
    """Returns a language specific Mnemonic instance. If the language code is not supported, English is used."""
    if language_code not in MNEMONIC_LANGUAGES:
        logger.debug(
            f"Language code '{language_code}' not supported by python-mnemonic; using English"
        )
        language_code = "en"
    return Mnemonic(language=MNEMONIC_LANGUAGES[language_code])


def generate_mnemonic(strength=256, language_code="en") -> str:
    """returns a 24 wordlist:
    "tomorrow question cook lend burden bone own junior stage square leaf father edge decrease pipe tired useful junior calm silver topple require rug clock"

    :param strength: 256 (default) will give you 24 words. 128 will result in 12 words
    :param language_code: only "en", "es", and "it" are supported. If you use an unsupported one, it'll fallback to English.
    """
    mnemo = initialize_mnemonic(language_code)
    return mnemo.generate(strength=strength)


def get_language(mnemonic: str) -> str:
    """Returns the language based on the mnemonic provided, errors are raised if a language is not supported or not properly detected"""
    try:
        supported_language_found = False
        language = Mnemonic.detect_language(mnemonic)
        words_as_list = mnemonic.split()
        # If we get French as language we have to double check whether it is not really an English mnemonic because of overlaps in the wordlists
        if language == "french":
            count = 0
            for word in words_as_list:
                if word in WORDLIST:
                    count += 1
            if count == 12:
                language = "english"
        for key, value in MNEMONIC_LANGUAGES.items():
            if value == language:
                supported_language_found = True
                return language
        if supported_language_found == False:
            raise SpecterError(
                f"The language {language.capitalize()} is not supported."
            )
    except ConfigurationError as ce:
        raise SpecterError(str(ce))


def validate_mnemonic(mnemonic: str) -> bool:
    """Returns true if the length and the checksum of the mnemonic are correct"""
    language = get_language(mnemonic)
    mnemo = Mnemonic(language)
    return mnemo.check(mnemonic)


def get_wordlist(mnemonic: str) -> list:
    """Returns a language specific wordlist"""
    language = get_language(mnemonic)
    mnemo = Mnemonic(language)
    wordlist = mnemo.wordlist
    return wordlist


def mnemonic_to_root(mnemonic: str, passphrase: str) -> bip32.HDKey:
    """Returns an embit BIP32 root key class"""
    wordlist = get_wordlist(mnemonic)
    seed = bip39.mnemonic_to_seed(mnemonic, passphrase, wordlist=wordlist)
    root = bip32.HDKey.from_seed(seed)
    return root



import logging
import random
from binascii import unhexlify

from flask import Blueprint
from flask import current_app as app
from flask import flash, make_response, redirect, render_template, request, url_for
from flask_babel import lazy_gettext as _
from flask_login import login_required

from ..helpers import notify_upgrade
from ..managers.wallet_manager import purposes

logger = logging.getLogger(__name__)

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
welcome_endpoint = Blueprint("welcome_endpoint", __name__)


@welcome_endpoint.route("/")
@login_required
def index():
    if request.args.get("mode"):
        if request.args.get("mode") == "remote":
            pass
    notify_upgrade(app, flash)
    if len(app.specter.wallet_manager.wallets) > 0:
        if len(app.specter.wallet_manager.wallets) > 1:
            return redirect(url_for("wallets_endpoint.wallets_overview"))
        return redirect(
            url_for(
                "wallets_endpoint.wallet",
                wallet_alias=app.specter.wallet_manager.wallets[
                    app.specter.wallet_manager.wallets_names[0]
                ].alias,
            )
        )

    return redirect(url_for("welcome_endpoint.about"))


@welcome_endpoint.route("/about", methods=["GET", "POST"])
@login_required
def about():
    notify_upgrade(app, flash)
    if request.method == "POST":
        action = request.form["action"]
        if action == "cancelsetup":
            app.specter.setup_status["stage"] = "start"
            app.specter.reset_setup("bitcoind")
            app.specter.reset_setup("torbrowser")

    return render_template(
        "base.jinja",
        specter=app.specter,
        rand=rand,
        supported_languages=app.supported_languages,
    )


@welcome_endpoint.route("/bitcoin.pdf")
@login_required
def get_whitepaper():
    if app.specter.chain == "main":
        if not app.specter.info["pruned"]:
            raw_tx = app.specter.rpc.getrawtransaction(
                "54e48e5f5c656b26c3bca14a8c95aa583d07ebe84dde3b7dd4a78f4e4186e713",
                False,
                "00000000000000ecbbff6bafb7efa2f7df05b227d5c73dca8f2635af32a2e949",
            )
            outputs = raw_tx.split("0100000000000000")
            pdf = ""
            for output in outputs[1:-2]:
                cur = 6
                pdf += output[cur : cur + 130]
                cur += 132
                pdf += output[cur : cur + 130]
                cur += 132
                pdf += output[cur : cur + 130]
            pdf += outputs[-2][6:-4]
        else:
            outputs_prun = app.specter.rpc.multi(
                [
                    (
                        "gettxout",
                        "54e48e5f5c656b26c3bca14a8c95aa583d07ebe84dde3b7dd4a78f4e4186e713",
                        i,
                    )
                    for i in range(0, 946)
                ]
            )
            pdf = ""
            for output in outputs_prun[:-1]:
                cur = 4
                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
                cur += 132
                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
                cur += 132
                pdf += output["result"]["scriptPubKey"]["hex"][cur : cur + 130]
            pdf += outputs_prun[-1]["result"]["scriptPubKey"]["hex"][4:-4]
        res = make_response(unhexlify(pdf[16:-16]))
        res.headers.set("Content-Disposition", "attachment")
        res.headers.set("Content-Type", "application/pdf")
        return res
    else:
        return render_template(
            "500.jinja",
            error=_(
                "You need a mainnet node to retrieve the whitepaper. Check your node configurations."
            ),
        )



import hashlib


def sha256sum(filename):
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    return h.hexdigest()



from ..device import Device
import hwilib.commands as hwi_commands
import importlib


class HWIDevice(Device):

    hwi_support = True
    exportable_to_wallet = False

    def create_psbts(self, base64_psbt, wallet):
        return {"hwi": wallet.fill_psbt(base64_psbt)}

    @classmethod
    def enumerate(cls, *args, **kwargs):
        mod = importlib.import_module(f"hwilib.devices.{cls.device_type}")
        return mod.enumerate(*args, **kwargs)

    @classmethod
    def get_client(cls, *args, **kwargs):
        return hwi_commands.get_client(cls.device_type, *args, **kwargs)



import os, time, requests, secrets, platform, tarfile, zipfile, sys, shutil
from ..process_controller.bitcoind_controller import BitcoindPlainController
import pgpy
from pathlib import Path
from .sha256sum import sha256sum
import logging
from .file_download import download_file
from ..specter_error import handle_exception, ExtProcTimeoutException
from .rpcauth import generate_salt, password_to_hmac

logger = logging.getLogger(__name__)


def setup_bitcoind_thread(specter=None, internal_bitcoind_version=""):
    specter.update_setup_status("bitcoind", "STARTING_SETUP")
    try:
        BITCOIND_OS_SUFFIX = {
            "Windows": "win64.zip",
            "Linux": "x86_64-linux-gnu.tar.gz",
            "Darwin": "osx64.tar.gz",
        }
        # ARM Linux devices (e.g. Raspberry Pi 4 == armv7l) need ARM binary
        if platform.system() == "Linux" and "armv" in platform.machine():
            BITCOIND_OS_SUFFIX["Linux"] = "arm-linux-gnueabihf.tar.gz"

        packed_name = (
            os.path.join(
                sys._MEIPASS,
                f"bitcoind/bitcoin-{internal_bitcoind_version}-{BITCOIND_OS_SUFFIX[platform.system()]}",
            )
            if getattr(sys, "frozen", False)
            else Path(__file__).parent
            / f"../../../../pyinstaller/bitcoind/bitcoin-{internal_bitcoind_version}-{BITCOIND_OS_SUFFIX[platform.system()]}"
        )
        bitcoin_binaries_folder = os.path.join(specter.data_folder, "bitcoin-binaries")
        logger.info(f"Unpacking binaries to {bitcoin_binaries_folder}")
        if BITCOIND_OS_SUFFIX[platform.system()].endswith("tar.gz"):
            with tarfile.open(packed_name, "r:gz") as so:
                so.extractall(specter.data_folder)
        else:
            with zipfile.ZipFile(packed_name, "r") as zip_ref:
                zip_ref.extractall(specter.data_folder)
        if os.path.exists(bitcoin_binaries_folder):
            shutil.rmtree(bitcoin_binaries_folder)
        os.rename(
            os.path.join(specter.data_folder, f"bitcoin-{internal_bitcoind_version}"),
            bitcoin_binaries_folder,
        )
        specter.reset_setup("bitcoind")
    except Exception as e:
        logger.error(f"Failed to install Bitcoin Core. Error: {e}")
        handle_exception(e)
        specter.update_setup_error("bitcoind", str(e))


def setup_bitcoind_directory_thread(
    specter=None, quicksync=True, pruned=True, node_alias=""
):
    try:
        if quicksync:
            prunednode_file = os.path.join(
                os.path.join(specter.data_folder, "snapshot-prunednode.zip")
            )
            prunednode_sha256sums_file = os.path.join(
                specter.data_folder, "prunednode-sha256sums.asc"
            )
            logger.info(f"Downloading latest.zip to {prunednode_file}")
            download_file(
                specter,
                "https://prunednode.today/latest.zip",
                prunednode_file,
                "bitcoind",
                "Downloading QuickSync files...",
            )
            logger.info(
                f"Downloading latest.signed.txt to {prunednode_sha256sums_file}"
            )
            download_file(
                specter,
                "https://prunednode.today/latest.signed.txt",
                prunednode_sha256sums_file,
                "bitcoind",
                "Downloading Quicksync signature...",
            )
            specter.update_setup_status("bitcoind", "VERIFY_SIGS")
            logger.info(f"Verifying signatures of {prunednode_sha256sums_file}")
            with open(prunednode_sha256sums_file, "r") as f:
                signed_sums = f.read()
                prunednode_release_pgp_key, _ = pgpy.PGPKey.from_file(
                    os.path.join(
                        sys._MEIPASS, "static/pruned-node-today-release-pubkey.asc"
                    )
                    if getattr(sys, "frozen", False)
                    else Path(__file__).parent
                    / "../static/pruned-node-today-release-pubkey.asc"
                )
                prunednode_sha256sums_msg = pgpy.PGPMessage.from_file(
                    prunednode_sha256sums_file
                )
                if not prunednode_release_pgp_key.verify(prunednode_sha256sums_msg):
                    raise Exception("Failed to verify prunednode.today PGP signature")
                prunednode_hash = sha256sum(prunednode_file)
                if prunednode_hash not in signed_sums:
                    raise Exception(
                        "Failed to verify prunednode.today hash is in SHA265SUMS.asc"
                    )
                logger.info(
                    f"Unpacking {prunednode_file} to {os.path.expanduser(specter.node_manager.get_by_alias(node_alias).datadir)}"
                )
                with zipfile.ZipFile(prunednode_file, "r") as zip_ref:
                    zip_ref.extractall(
                        os.path.expanduser(
                            specter.node_manager.get_by_alias(node_alias).datadir
                        )
                    )
                os.remove(prunednode_file)

        logger.info(f"Writing bitcoin.conf")
        if not os.path.exists(specter.node_manager.get_by_alias(node_alias).datadir):
            os.makedirs(specter.node_manager.get_by_alias(node_alias).datadir)
        with open(
            os.path.join(
                specter.node_manager.get_by_alias(node_alias).datadir, "bitcoin.conf"
            ),
            "w+",
        ) as file:
            salt = generate_salt(16)
            password_hmac = password_to_hmac(
                salt, specter.node_manager.get_by_alias(node_alias).password
            )
            file.write(
                f"\nrpcauth={specter.node_manager.get_by_alias(node_alias).user}:{salt}${password_hmac}"
            )
            file.write(f"\nserver=1")
            file.write(f"\nlisten=1")
            file.write(f"\ndisablewallet=0")
            file.write(f"\nonion=127.0.0.1:9050")
            file.write(f"\nbind=127.0.0.1")
            file.write(f"\ntorcontrol=127.0.0.1:9051")
            file.write(f"\ntorpassword={specter.config['torrc_password']}")
            file.write(f"\nfallbackfee=0.0002")
            if quicksync or pruned:
                file.write(f"\nprune=1000")
            else:
                file.write(f"\nblockfilterindex=1")
            file.write(f"\n[test]")
            file.write(f"\nbind=127.0.0.1")
            file.write(f"\n[regtest]")
            file.write(f"\nbind=127.0.0.1")
            file.write(f"\n[signet]")
            file.write(f"\nbind=127.0.0.1")

        specter.update_setup_status("bitcoind", "START_SERVICE")

        # Specter's 'bitcoind' attribute will instantiate a BitcoindController as needed
        logger.info(
            f"Starting up Bitcoin Core... in {os.path.expanduser(specter.node_manager.get_by_alias(node_alias).datadir)}"
        )
        success = specter.node_manager.get_by_alias(node_alias).start(timeout=60)
        specter.update_active_node(specter.node_manager.get_by_alias(node_alias).alias)
        if not success:
            specter.update_setup_status("bitcoind", "FAILED")
            logger.info("No success connecting to Bitcoin Core")
        specter.check()
        specter.reset_setup("bitcoind")
        specter.setup_status["stage"] = "end"
    except ExtProcTimeoutException as e:
        e.check_logfile(
            os.path.join(
                specter.node_manager.get_by_alias(node_alias).datadir, "debug.log"
            )
        )
        logger.error(f"Failed to setup Bitcoin Core. Error: {e}")
        logger.error(e.get_logger_friendly())
        specter.update_setup_error("bitcoind", str(e))
    except Exception as e:
        logger.exception(f"Failed to setup Bitcoin Core. Error: {e}")
        specter.update_setup_error("bitcoind", str(e))



import os

from cryptoadvance.specter.managers.genericdata_manager import GenericDataManager
from cryptoadvance.specter.wallet import Wallet


class ServiceAnnotationsStorage(GenericDataManager):
    """
    Stores Service-specific annotations for addresses and txs. Each wallet + service
    pairing will get its own <wallet_alias>_<service>.json file.

    Annotations must be in a json-serializable dict, keyed on addr or tx:
    {
        "someaddress": {
            "foo": 1,
            "bar": "hello",
        }
    }
    """

    def __init__(self, service_id: str, wallet: Wallet):
        # Must set these before calling parent's __init__() so that the data_file
        #   property will have the proper member vars.
        self.service_id = service_id
        self.wallet = wallet

        super().__init__(wallet.manager.data_folder)

        if "addrs" not in self.data:
            self.data["addrs"] = {}

        if "txs" not in self.data:
            self.data["txs"] = {}

    @property
    def data_file(self):
        # TODO: currently saving in the wallets/ dir but not in the main vs regtest subdir
        return os.path.join(
            self.wallet.manager.data_folder,
            f"{self.wallet.alias}_{self.service_id}.json",
        )

    def save(self):
        # Expose for external use
        self._save()

    def set_addr_annotations(self, addr: str, annotations: dict, autosave: bool = True):
        self.data["addrs"][addr] = annotations
        if autosave:
            self._save()

    def remove_addr_annotations(self, addr: str, autosave: bool = True):
        self.data["addrs"].pop(addr, None)
        if autosave:
            self._save()

    def get_addr_annotations(self, addr: str):
        return self.data["addrs"].get(addr, None)

    def get_all_addr_annotations(self):
        return self.data["addrs"]

    def set_tx_annotations(self, txid: str, annotations: dict, autosave: bool = True):
        self.data["txs"][txid] = annotations
        if autosave:
            self._save()

    def get_tx_annotations(self, txid: str):
        return self.data["txs"].get(txid, None)

    def get_all_tx_annotations(self):
        return self.data["txs"]



import serial
import logging


logger = logging.getLogger("jade.serial")


#
# Low-level Serial backend interface to Jade
# Calls to send and receive bytes over the interface.
# Intended for use via JadeInterface wrapper.
#
# Either:
#  a) use via JadeInterface.create_serial() (see JadeInterface)
# (recommended)
# or:
#  b) use JadeSerialImpl() directly, and call connect() before
#     using, and disconnect() when finished,
# (caveat cranium)
#
class JadeSerialImpl:
    def __init__(self, device, baud, timeout):
        self.device = device
        self.baud = baud
        self.timeout = timeout
        self.ser = None

    def connect(self):
        assert self.ser is None

        logger.info("Connecting to {} at {}".format(self.device, self.baud))
        self.ser = serial.Serial(
            self.device, self.baud, timeout=self.timeout, write_timeout=self.timeout
        )
        assert self.ser is not None
        self.ser.__enter__()
        logger.info("Connected")

    def disconnect(self):
        assert self.ser is not None
        self.ser.__exit__()

        # Reset state
        self.ser = None

    def write(self, bytes_):
        assert self.ser is not None
        return self.ser.write(bytes_)

    def read(self, n):
        assert self.ser is not None
        return self.ser.read(n)



import cryptography
import json
import logging
import os
import platform
import random
import secrets
import shutil
import sys
import tarfile
import time
import zipfile
from pathlib import Path

import pgpy
import requests
from flask import Blueprint, Flask
from flask import current_app as app
from flask import flash, jsonify, redirect, render_template, request, send_file, url_for
from flask_babel import lazy_gettext as _
from flask_login import current_user, login_required
from cryptoadvance.specter.services.service import Service

from cryptoadvance.specter.user import User, UserSecretException

from ..helpers import (
    get_loglevel,
    get_startblock_by_chain,
    notify_upgrade,
    set_loglevel,
)
from ..persistence import write_devices, write_wallet
from ..specter_error import ExtProcTimeoutException, handle_exception
from ..util.sha256sum import sha256sum
from ..util.shell import get_last_lines_from_file
from ..util.tor import start_hidden_service, stop_hidden_services

logger = logging.getLogger(__name__)

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
settings_endpoint = Blueprint("settings_endpoint", __name__)


@settings_endpoint.route("/", methods=["GET"])
@login_required
def settings():
    return redirect(url_for("settings_endpoint.general"))


@settings_endpoint.route("/general", methods=["GET", "POST"])
@login_required
def general():
    current_version = notify_upgrade(app, flash)
    explorer_id = app.specter.explorer_id
    fee_estimator = app.specter.fee_estimator
    fee_estimator_custom_url = app.specter.config.get("fee_estimator_custom_url", "")
    loglevel = get_loglevel(app)
    unit = app.specter.unit
    services = app.specter.service_manager.services
    if request.method == "POST":
        action = request.form["action"]

        autohide_sensitive_info_timeout = request.form[
            "autohide_sensitive_info_timeout"
        ]
        if autohide_sensitive_info_timeout == "NEVER":
            autohide_sensitive_info_timeout = None
        elif autohide_sensitive_info_timeout == "CUSTOM":
            autohide_sensitive_info_timeout = int(
                request.form["custom_autohide_sensitive_info_timeout"]
            )
        else:
            autohide_sensitive_info_timeout = int(autohide_sensitive_info_timeout)

        if "autologout_timeout" in request.form:
            # Is only in the form if specter.config.auth.method != "none"
            autologout_timeout = request.form["autologout_timeout"]
            if autologout_timeout == "NEVER":
                autologout_timeout = None
            elif autologout_timeout == "CUSTOM":
                autologout_timeout = int(request.form["custom_autologout_timeout"])
            else:
                autologout_timeout = int(autologout_timeout)
        else:
            autologout_timeout = None

        explorer_id = request.form["explorer"]
        explorer_data = app.config["EXPLORERS_LIST"][explorer_id]
        if explorer_id == "CUSTOM":
            explorer_data["url"] = request.form["custom_explorer"]
        fee_estimator = request.form["fee_estimator"]
        fee_estimator_custom_url = request.form["fee_estimator_custom_url"]
        unit = request.form["unit"]
        validate_merkleproof_bool = request.form.get("validatemerkleproof") == "on"
        if current_user.is_admin:
            active_services = []
            for service_name in services:
                if request.form.get(f"service_{service_name}"):
                    active_services.append(service_name)

            loglevel = request.form["loglevel"]

        if action == "save":
            if current_user.is_admin:
                set_loglevel(app, loglevel)

            app.specter.config_manager.update_autohide_sensitive_info_timeout(
                autohide_sensitive_info_timeout, current_user
            )
            app.specter.config_manager.update_autologout_timeout(
                autologout_timeout, current_user
            )

            app.specter.update_explorer(explorer_id, explorer_data, current_user)
            app.specter.update_unit(unit, current_user)
            app.specter.update_merkleproof_settings(
                validate_bool=validate_merkleproof_bool
            )
            if current_user.is_admin:
                app.specter.service_manager.set_active_services(active_services)
            app.specter.update_fee_estimator(
                fee_estimator=fee_estimator,
                custom_url=fee_estimator_custom_url,
                user=current_user,
            )
            app.specter.check()

        elif action == "restore":
            restore_devices = []
            restore_wallets = []
            if request.form.get("restoredevices", ""):
                restore_devices = json.loads(request.form.get("restoredevices", "[]"))
            if request.form.get("restorewallets", ""):
                restore_wallets = json.loads(request.form.get("restorewallets", "[]"))
            write_devices(restore_devices)
            app.specter.device_manager.update()

            rescanning = False
            for wallet in restore_wallets:
                try:
                    app.specter.wallet_manager.rpc.createwallet(
                        os.path.join(
                            app.specter.wallet_manager.rpc_path, wallet["alias"]
                        ),
                        True,
                    )
                except Exception as e:
                    # if wallet already exists in Bitcoin Core
                    # continue with the existing one
                    if "already exists" not in str(e):
                        flash(
                            _("Failed to import wallet {}, error: {}").format(
                                wallet["name"], e
                            ),
                            "error",
                        )
                        handle_exception(e)
                        continue
                    logger.debug(
                        f"Wallet {wallet['alias']} already exists, skipping creation"
                    )
                write_wallet(wallet)
                app.specter.wallet_manager.update(use_threading=False)

                try:
                    wallet_obj = app.specter.wallet_manager.get_by_alias(
                        wallet["alias"]
                    )
                    wallet_obj.keypoolrefill(0, wallet_obj.IMPORT_KEYPOOL, change=False)
                    wallet_obj.keypoolrefill(0, wallet_obj.IMPORT_KEYPOOL, change=True)
                    wallet_obj.import_labels(wallet.get("labels", {}))
                    try:
                        wallet_obj.rpc.rescanblockchain(
                            wallet["blockheight"]
                            if "blockheight" in wallet
                            else get_startblock_by_chain(app.specter),
                            no_wait=True,
                        )
                        app.logger.info("Rescanning Blockchain ...")
                        rescanning = True
                    except Exception as e:
                        app.logger.error(
                            "Exception while rescanning blockchain for wallet {}: {}".format(
                                wallet["alias"], e
                            )
                        )
                        flash(
                            _("Failed to perform rescan for wallet: {}").format(e),
                            "error",
                        )
                    wallet_obj.getdata()
                except Exception as e:
                    flash(
                        _("Failed to import wallet {}").format(wallet["name"]), "error"
                    )
                    handle_exception(e)
            flash(_("Specter data was successfully loaded from backup"), "info")
            if rescanning:
                flash(
                    _(
                        "Wallets are rescanning for transactions history.\n\
This may take a few hours to complete."
                    ),
                    "info",
                )

    return render_template(
        "settings/general_settings.jinja",
        services=services,
        fee_estimator=fee_estimator,
        fee_estimator_custom_url=fee_estimator_custom_url,
        loglevel=loglevel,
        validate_merkle_proofs=app.specter.config.get("validate_merkle_proofs") is True,
        unit=unit,
        specter=app.specter,
        current_version=current_version,
        rand=rand,
        supported_languages=app.supported_languages,
    )


@settings_endpoint.route("/tor", methods=["GET", "POST"])
@login_required
def tor():
    """
    controls the tor related settings
    GET for displaying the page, POST for updates
    param action might be "save", "test_tor" or "toggle_hidden_service"
    param proxy_url the Tor deamon url, usually something like socks5h://localhost:9050
    param only_tor "on" or something else ("off")
    """
    if not current_user.is_admin:
        flash(_("Only an admin is allowed to access this page."), "error")
        return redirect("")
    app.specter.reset_setup("torbrowser")
    current_version = notify_upgrade(app, flash)
    proxy_url = app.specter.proxy_url
    only_tor = app.specter.only_tor
    tor_control_port = app.specter.tor_control_port
    tor_type = app.specter.tor_type
    if request.method == "POST":
        action = request.form["action"]
        tor_type = request.form["tor_type"]
        proxy_url = request.form["proxy_url"]
        only_tor = request.form.get("only_tor") == "on"
        tor_control_port = request.form["tor_control_port"]
        hidden_service = request.form.get("hidden_service") == "on"

        if action == "save":
            logger.info("Updating Tor settings...")
            app.specter.update_tor_type(tor_type, current_user)

            if tor_type == "custom":
                app.specter.update_proxy_url(proxy_url, current_user)
                app.specter.update_tor_control_port(tor_control_port, current_user)
            else:
                proxy_url = "socks5h://localhost:9050"
                tor_control_port = ""

            app.specter.update_only_tor(only_tor, current_user)
            if hidden_service != app.specter.config["tor_status"]:
                if not app.config["DEBUG"]:
                    if app.specter.config["auth"].get("method", "none") == "none":
                        flash(
                            "Enabling Tor hidden service will expose your Specter for remote access.<br>It is therefore required that you set up authentication tab for Specter first to prevent unauthorized access.<br><br>Please go to Settings -> Authentication and set up an authentication method and retry.",
                            "error",
                        )
                    else:
                        if hasattr(current_user, "is_admin") and current_user.is_admin:
                            if not hidden_service:
                                stop_hidden_services(app)
                                app.specter.toggle_tor_status()
                                flash(
                                    "Tor hidden service turn off successfully", "info"
                                )
                            else:
                                try:
                                    start_hidden_service(app)
                                    app.specter.toggle_tor_status()
                                    flash(
                                        "Tor hidden service turn on successfully",
                                        "info",
                                    )
                                except Exception as e:
                                    handle_exception(e)
                                    flash(
                                        "Failed to start Tor hidden service. Make sure you have Tor running with ControlPort configured and try again. Error returned: {}".format(
                                            e
                                        ),
                                        "error",
                                    )
                else:
                    flash(
                        "Can't toggle hidden service while Specter is running in DEBUG mode",
                        "error",
                    )

            app.specter.check()

        elif action == "starttor":
            logger.info("Starting Tor...")
            try:
                app.specter.tor_daemon.start_tor_daemon()
                flash(_("Specter has started Tor"))
            except Exception as e:
                flash(_("Failed to start Tor, error: {}").format(e), "error")
                logger.error(f"Failed to start Tor, error: {e}")
        elif action == "stoptor":
            logger.info("Stopping Tor...")
            try:
                app.specter.tor_daemon.stop_tor_daemon()
                time.sleep(1)
                flash(_("Specter stopped Tor successfully"))
            except Exception as e:
                flash(_("Failed to stop Tor, error: {}").format(e), "error")
                logger.error(f"Failed to start Tor, error: {e}")
        elif action == "uninstalltor":
            logger.info("Uninstalling Tor...")
            try:
                if app.specter.is_tor_dameon_running():
                    app.specter.tor_daemon.stop_tor_daemon()
                shutil.rmtree(os.path.join(app.specter.data_folder, "tor-binaries"))
                os.remove(os.path.join(app.specter.data_folder, "torrc"))
                flash(_("Tor uninstalled successfully"))
            except Exception as e:
                flash(_("Failed to uninstall Tor, error: {}").format(e), "error")
                logger.error(f"Failed to uninstall Tor, error: {e}")
        elif action == "test_tor":
            logger.info("Testing the Tor connection...")
            try:
                requests_session = requests.Session()
                requests_session.proxies["http"] = proxy_url
                requests_session.proxies["https"] = proxy_url
                res = requests_session.get(
                    "http://2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion/",  # Tor Project onion v3 website
                    timeout=30,
                )
                tor_connectable = res.status_code == 200
                if tor_connectable:
                    flash(_("Tor requests test completed successfully!"), "info")
                else:
                    flash(
                        _(
                            "Failed to make test request over Tor. Status-Code: {}"
                        ).format(res.status_code),
                        "error",
                    )
                    logger.error(
                        f"Failed to make test request over Tor. Status-Code: {res.status_code}"
                    )
                    if tor_type == "builtin":
                        logger.error("Tor-Logs:")
                        app.specter.tor_daemon.stop_tor_daemon()
                        time.sleep(1)
                        logger.error(app.specter.tor_daemon.get_logs())
                        app.specter.tor_daemon.start_tor_daemon()
            except Exception as e:
                flash(
                    _("Failed to make test request over Tor.\nError: {}").format(e),
                    "error",
                )
                logger.error(f"Failed to make test request over Tor.\nError: {e}")
                if tor_type == "builtin":
                    logger.error("Tor-Logs:")
                    app.specter.tor_daemon.stop_tor_daemon()
                    time.sleep(1)
                    logger.error(app.specter.tor_daemon.get_logs())
                    app.specter.tor_daemon.start_tor_daemon()
                tor_connectable = False

    return render_template(
        "settings/tor_settings.jinja",
        tor_type=tor_type,
        proxy_url=proxy_url,
        only_tor=only_tor,
        tor_control_port=tor_control_port,
        tor_service_id=app.tor_service_id,
        torbrowser_installed=os.path.isfile(app.specter.torbrowser_path),
        torbrowser_running=app.specter.is_tor_dameon_running(),
        specter=app.specter,
        current_version=current_version,
        rand=rand,
    )


@settings_endpoint.route("/auth", methods=["GET", "POST"])
@login_required
def auth():
    # TODO: Simplify this endpoint. Separate out setting the Authentication mode from all
    # the other options here: updating admin username/password, adding users, deleting
    # users, etc. Do those in simple separate screens with their own endpoints.
    current_version = notify_upgrade(app, flash)
    auth = app.specter.config["auth"]
    method = auth["method"]
    rate_limit = auth["rate_limit"]
    registration_link_timeout = auth["registration_link_timeout"]
    users = None

    if (
        request.method == "GET"
        and current_user.is_admin
        and app.config["LOGIN_DISABLED"] == False
        and not current_user.plaintext_user_secret
    ):
        # Password protection is enabled but the admin user doesn't have their
        # user_secret decrypted. Force them to login again to prevent problems if they
        # try to change their password (changing password requires re-encrypting the
        # user_secret... which we can't do if it isn't decrypted first).
        flash(_("Must login again to before making Authentication changes"))
        return redirect(
            url_for("auth_endpoint.login")
            + "?next="
            + url_for("settings_endpoint.auth")
        )

    if current_user.is_admin and method == "usernamepassword":
        users = [user for user in app.specter.user_manager.users if not user.is_admin]
    if request.method == "POST":
        action = request.form["action"]

        if action == "save":
            if "specter_username" in request.form:
                specter_username = request.form["specter_username"]
                specter_password = request.form["specter_password"]
            else:
                specter_username = None
                specter_password = None

            if current_user.is_admin:
                method = request.form["method"]
                rate_limit = request.form["rate_limit"]
                registration_link_timeout = request.form["registration_link_timeout"]

            min_chars = int(auth["password_min_chars"])
            if specter_username:

                if current_user.username != specter_username:
                    if app.specter.user_manager.get_user_by_username(specter_username):
                        flash(
                            _("Username is already taken, please choose another one"),
                            "error",
                        )
                        return render_template(
                            "settings/auth_settings.jinja",
                            method=method,
                            rate_limit=rate_limit,
                            registration_link_timeout=registration_link_timeout,
                            users=users,
                            specter=app.specter,
                            current_version=current_version,
                            rand=rand,
                            has_service_encrypted_storage=app.specter.service_manager.user_has_encrypted_storage(
                                current_user
                            ),
                        )
                    current_user.username = specter_username

                if specter_password:
                    if len(specter_password) < min_chars:
                        flash(
                            _(
                                "Please enter a password of a least {} characters"
                            ).format(min_chars),
                            "error",
                        )
                        return render_template(
                            "settings/auth_settings.jinja",
                            method=method,
                            rate_limit=rate_limit,
                            registration_link_timeout=registration_link_timeout,
                            users=users,
                            specter=app.specter,
                            current_version=current_version,
                            rand=rand,
                            has_service_encrypted_storage=app.specter.service_manager.user_has_encrypted_storage(
                                current_user
                            ),
                        )
                    current_user.set_password(specter_password)

                current_user.save_info()

            if current_user.is_admin:
                app.specter.update_auth(method, rate_limit, registration_link_timeout)
                if method in ["rpcpasswordaspin", "passwordonly", "usernamepassword"]:
                    if method == "passwordonly":
                        specter_password = request.form.get("specter_password_only")
                        if specter_password and len(specter_password) < min_chars:
                            flash(
                                _(
                                    "Please enter a password of a least {} characters"
                                ).format(min_chars),
                                "error",
                            )
                            return render_template(
                                "settings/auth_settings.jinja",
                                method=method,
                                rate_limit=rate_limit,
                                registration_link_timeout=registration_link_timeout,
                                users=users,
                                specter=app.specter,
                                current_version=current_version,
                                rand=rand,
                                has_service_encrypted_storage=app.specter.service_manager.user_has_encrypted_storage(
                                    current_user
                                ),
                            )
                        elif not specter_password:
                            # Set to the default
                            specter_password = "admin"

                        try:
                            current_user.set_password(specter_password)
                        except UserSecretException as e:
                            # Most likely the admin User is logged in but the
                            # plaintext_user_secret is not available in memory (happens
                            # when the server restarts).
                            logger.warn(e)
                            flash(
                                _(
                                    "Error re-encrypting Service data. Log out and log back in before trying again."
                                ),
                                "error",
                            )
                            return render_template(
                                "settings/auth_settings.jinja",
                                method=method,
                                rate_limit=rate_limit,
                                registration_link_timeout=registration_link_timeout,
                                users=users,
                                specter=app.specter,
                                current_version=current_version,
                                rand=rand,
                                has_service_encrypted_storage=app.specter.service_manager.user_has_encrypted_storage(
                                    current_user
                                ),
                            )

                        current_user.save_info()

                        flash(
                            _("Admin password successfully updated"),
                            "info",
                        )

                    if method == "usernamepassword":
                        users = [
                            user
                            for user in app.specter.user_manager.users
                            if not user.is_admin
                        ]
                    else:
                        users = None

                    app.config["LOGIN_DISABLED"] = False
                else:
                    users = None
                    app.config["LOGIN_DISABLED"] = True

                    # Cannot support Services if there's no password (admin was already
                    # warned about this in the UI). Remove User.services, clear the
                    # `user_secret`, and wipe the ServiceEncryptedStorage.
                    app.specter.service_manager.remove_all_services_from_user(
                        current_user
                    )

            # Redirect if a URL was given via the next variable
            if request.form.get("next") and request.form.get("next") != "":
                return redirect(request.form.get("next"))
            app.specter.check()

        elif action == "adduser":
            if current_user.is_admin:
                new_otp = secrets.token_urlsafe(16)
                now = time.time()
                timeout = int(registration_link_timeout)
                timeout = 0 if timeout < 0 else timeout
                if timeout > 0:
                    expiry = now + timeout * 60 * 60
                    if timeout > 1:
                        expiry_desc = " " + _("(expires in {} hours)").format(timeout)
                    else:
                        expiry_desc = " " + _("(expires in 1 hour)")
                else:
                    expiry = 0
                    expiry_desc = ""
                app.specter.otp_manager.add_new_user_otp(
                    {"otp": new_otp, "created_at": now, "expiry": expiry}
                )
                flash(
                    _("New user link generated{}: {}auth/register?otp={}").format(
                        expiry_desc, request.url_root, new_otp
                    ),
                    "info",
                )
            else:
                flash(
                    _("Error: Only the admin account can issue new registration links"),
                    "error",
                )

        elif action == "deleteuser":
            delete_user = request.form["deleteuser"]
            user = app.specter.user_manager.get_user(delete_user)
            if current_user.is_admin:
                # TODO: delete should be done by UserManager
                app.specter.delete_user(user)
                users.remove(user)
                flash(
                    _("User {} was deleted successfully").format(user.username), "info"
                )
            else:
                flash(_("Error: Only the admin account can delete users"), "error")

    return render_template(
        "settings/auth_settings.jinja",
        method=method,
        rate_limit=rate_limit,
        registration_link_timeout=registration_link_timeout,
        users=users,
        specter=app.specter,
        current_version=current_version,
        rand=rand,
        has_service_encrypted_storage=app.specter.service_manager.user_has_encrypted_storage(
            current_user
        ),
        next=request.args.get("next", ""),
    )


@settings_endpoint.route("/hwi", methods=["GET", "POST"])
@login_required
def hwi():
    current_version = notify_upgrade(app, flash)
    if request.method == "POST":
        hwi_bridge_url = request.form["hwi_bridge_url"]
        app.specter.update_hwi_bridge_url(hwi_bridge_url, current_user)
        flash(_("HWIBridge URL is updated! Don't forget to whitelist Specter!"))
    return render_template(
        "settings/hwi_settings.jinja",
        specter=app.specter,
        current_version=current_version,
        rand=rand,
    )


@settings_endpoint.route("/assets/set_label", methods=["POST"])
@login_required
def set_asset_label():
    asset = request.form["asset"]
    label = request.form["label"].rstrip()
    if label.lower() in ["btc", "bitcoin", "sat", "lbtc", "tlbtc"]:
        return f'Label "{label}" is not allowed', 500
    try:
        app.specter.update_asset_label(asset, label)
    except Exception as e:
        return str(e), 500
    return {"success": True}


@settings_endpoint.route("/assets/", methods=["GET"])
@login_required
def assets():
    """List of all known / labeled assets"""
    return app.specter.asset_labels


@settings_endpoint.route("/assets/<asset>/", methods=["GET"])
@login_required
def get_asset(asset):
    """Get info about particular asset. Accepts either label or hex asset"""
    try:
        # check if we've got label, not an asset
        if len(asset) != 64:
            for a, lbl in app.specter.asset_labels.items():
                if lbl == asset:
                    return {"asset": a, "label": lbl}
            return {"error": "asset lookup by label failed"}, 404
        return {"asset": asset, "label": app.specter.asset_label(asset)}
    except Exception as e:
        return {"error": str(e)}, 500


################## Settings util endpoints #######################

# Specter backup file
@settings_endpoint.route("/specter_backup.zip")
@login_required
def backup_file():
    return send_file(
        app.specter.specter_backup_file(),
        attachment_filename="specter-backup.zip",
        as_attachment=True,
    )



import json
import logging

from decimal import Decimal
from urllib.parse import urlparse
from flask import redirect, render_template, request, url_for, flash
from flask import current_app as app
from flask.json import jsonify
from flask_babel import lazy_gettext as _
from flask_login import current_user, login_required
from functools import wraps
from cryptoadvance.specter.user import User
from cryptoadvance.specter.wallet import Wallet

from cryptoadvance.specter.services.service_encrypted_storage import (
    ServiceEncryptedStorageError,
)
from . import client as swan_client
from .client import SwanApiException
from .service import SwanService
from ..controller import user_secret_decrypted_required


logger = logging.getLogger(__name__)

swan_endpoint = SwanService.blueprint


def refreshtoken_required(func):
    """Refresh token needed for any endpoint that interacts with Swan API"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            if not SwanService.has_refresh_token():
                logger.debug(f"No refresh token, redirecting to relink Swan account")
                return redirect(
                    url_for(f"{SwanService.get_blueprint_name()}.oauth2_start")
                )
        except ServiceEncryptedStorageError as e:
            logger.debug(repr(e))
            flash("Re-login required to access your protected services data")

            # Use Flask's built-in re-login w/automatic redirect back to calling page
            return app.login_manager.unauthorized()
        return func(*args, **kwargs)

    return wrapper


@swan_endpoint.route("/")
@login_required
@user_secret_decrypted_required
def index():
    if SwanService.has_refresh_token():
        # User has already completed Swan integration; skip ahead
        return redirect(url_for(f"{SwanService.get_blueprint_name()}.withdrawals"))
    specter_used_hostname = urlparse(request.url).netloc
    return render_template(
        "swan/index.jinja",
        swan_frontend_url=app.config["SWAN_FRONTEND_URL"],
        specter_used_hostname=specter_used_hostname,
        # The oauth2-flow uses a whitelist. Is our hostname whitelisted?!
        specter_hostname_supported=specter_used_hostname
        in app.config["SWAN_ALLOWED_SPECTER_HOSTNAMES"],
        allowed_specter_hostnames=app.config["SWAN_ALLOWED_SPECTER_HOSTNAMES"],
    )


@swan_endpoint.route("/withdrawals")
@login_required
@refreshtoken_required
def withdrawals():
    # The wallet currently configured for ongoing autowithdrawals
    wallet: Wallet = SwanService.get_associated_wallet()

    return render_template(
        "swan/withdrawals.jinja",
        # txlist=swan_txs,
        wallet=wallet,
        services=app.specter.service_manager.services,
        swan_id=SwanService.id,
        autowithdrawal_threshold=SwanService.get_current_user_service_data().get(
            SwanService.AUTOWITHDRAWAL_THRESHOLD
        ),
    )


@swan_endpoint.route("/settings", methods=["GET"])
@login_required
@refreshtoken_required
def settings():
    associated_wallet: Wallet = SwanService.get_associated_wallet()

    # Get the user's Wallet objs, sorted by Wallet.name
    wallet_names = sorted(current_user.wallet_manager.wallets.keys())
    wallets = [current_user.wallet_manager.wallets[name] for name in wallet_names]

    return render_template(
        "swan/settings.jinja",
        associated_wallet=associated_wallet,
        wallets=wallets,
        cookies=request.cookies,
        num_reserved_addrs=SwanService.MIN_PENDING_AUTOWITHDRAWAL_ADDRS,
    )


@swan_endpoint.route("/settings/autowithdrawal", methods=["POST"])
@login_required
@refreshtoken_required
def update_autowithdrawal():
    threshold = request.form["threshold"]
    destination_wallet_alias = request.form["destination_wallet"]
    wallet = current_user.wallet_manager.get_by_alias(destination_wallet_alias)

    try:
        SwanService.set_autowithdrawal_settings(wallet=wallet, btc_threshold=threshold)
        return redirect(url_for(f"{SwanService.get_blueprint_name()}.withdrawals"))
    except SwanApiException as e:
        logger.exception(e)
        flash(_("Error communicating with Swan API"))
        return redirect(url_for(f"{SwanService.get_blueprint_name()}.settings"))


""" ***************************************************************************
                                OAuth2 endpoints
*************************************************************************** """


@swan_endpoint.route("/oauth2/start")
@login_required
@user_secret_decrypted_required
def oauth2_start():
    """
    Set up the Swan API integration by requesting our initial access_token and
    refresh_token.
    """
    # Do we have a token already?
    if SwanService.has_refresh_token():
        return redirect(url_for(f"{SwanService.get_blueprint_name()}.settings"))

    # Let's start the PKCE-flow
    specter_used_hostname = urlparse(request.url).netloc
    # The oauth2-flow uses a whitelist. Is our hostname whitelisted?!
    if specter_used_hostname not in app.config["SWAN_ALLOWED_SPECTER_HOSTNAMES"]:
        return redirect(url_for(f"{SwanService.get_blueprint_name()}.index"))
    flow_url = SwanService.client().get_oauth2_start_url(specter_used_hostname)

    return render_template(
        "swan/oauth2_start.jinja",
        flow_url=flow_url,
    )


@swan_endpoint.route("/integration_check")
def integration_check():
    """
    Polled by the oauth2_start page via AJAX. Returns False until we have a Swan refresh token.
    Returns True and can then redirect to settings.
    """
    try:
        if SwanService.has_refresh_token():
            return jsonify(success=True)
    except Exception as e:
        # Expected to fail in various possible ways: not logged in, user_secret
        # not decrypted, Swan integration not complete.
        pass
    return jsonify(success=False)


"""
    Note: the callback from Swan will be treated by Flask as an AnonymousUserMixin request
    but we need the user logged in and their user_secret decrypted. So we must require
    @login_required here which will interrupt the return flow with the login prompt.
"""


@swan_endpoint.route("/oauth2/callback")
@login_required
def oauth2_auth():
    if request.args.get("error"):
        logger.error(
            f"OAuth2 flow error: {request.args.get('error')}, {request.args.get('error_description')}"
        )
        return render_template(
            "500.jinja",
            error=request.args.get("error"),
            traceback=request.args.get(
                "error_description"
            ),  # Slightly misusing the traceback field
        )

    user: User = app.specter.user_manager.get_user()
    error = None

    try:
        SwanService.client().handle_oauth2_auth_callback(request)
        SwanService.store_new_api_access_data()
    except swan_client.SwanApiException as e:
        logger.exception(e)
        error = e

    if not error:
        try:
            # Add the Service to the User's profile (will now appear in sidebar)
            user = app.specter.user_manager.get_user()
            user.add_service(SwanService.id)

            # Sync this Specter instance with any previous Swan-Specter integrations
            SwanService.sync_swan_data()

            service_data = SwanService.get_current_user_service_data()
            if service_data.get(SwanService.SPECTER_WALLET_ALIAS):
                # We've re-synced with an existing auto-withdrawal setup. Redirect
                # straight to the auto-withdrawals page.
                return redirect(
                    url_for(f"{SwanService.get_blueprint_name()}.withdrawals")
                )

            return redirect(url_for(".oauth2_success"))
        except Exception as e:
            logger.exception(e)
            error = e

    return render_template(
        "500.jinja",
        response=None,
        error=str(error),
        error_description=_("Could not complete the OAuth callback from Swan"),
        cookies=request.cookies,
    )


@swan_endpoint.route("/oauth2/success")
def oauth2_success():
    """
    The redirect from the oauth2 callback has to land on an endpoint that does not
    have the @login_required filter set. Once we're back we can proceed to login-
    protected pages as usual.
    """
    return render_template(
        "swan/oauth2_success.jinja",
    )


@swan_endpoint.route("/oauth2/delete-token", methods=["POST"])
@login_required
@refreshtoken_required
def oauth2_delete_token():
    SwanService.remove_swan_integration(current_user)

    url = url_for(f"{SwanService.get_blueprint_name()}.index")
    return redirect(url_for(f"{SwanService.get_blueprint_name()}.index"))



import logging
from flask import Flask, Response, redirect, render_template, request, url_for
from flask import current_app as app
from flask_login import login_required, current_user

from cryptoadvance.specter.services.controller import user_secret_decrypted_required
from .service import DevhelpService
from cryptoadvance.specter.wallet import Wallet
from cryptoadvance.specter.util.common import robust_json_dumps

logger = logging.getLogger(__name__)

devhelp_endpoint = DevhelpService.blueprint


@devhelp_endpoint.route("/")
@login_required
@user_secret_decrypted_required
def index():
    return render_template(
        "devhelp/index.jinja",
    )


@devhelp_endpoint.route("/html/<html_component>")
@login_required
def html_component(html_component):
    associated_wallet: Wallet = DevhelpService.get_associated_wallet()
    return render_template(
        f"devhelp/html/{html_component}",
        wallet=associated_wallet,
        services=app.specter.service_manager.services,
        address=associated_wallet.get_address(3) if associated_wallet else None,
    )


@devhelp_endpoint.route("/console", methods=["GET"])
@login_required
def console():
    return render_template("devhelp/dev-console.jinja")


@devhelp_endpoint.route("/settings", methods=["GET"])
@login_required
@user_secret_decrypted_required
def settings_get():
    associated_wallet: Wallet = DevhelpService.get_associated_wallet()

    # Get the user's Wallet objs, sorted by Wallet.name
    wallet_names = sorted(current_user.wallet_manager.wallets.keys())
    wallets = [current_user.wallet_manager.wallets[name] for name in wallet_names]

    return render_template(
        "devhelp/settings.jinja",
        associated_wallet=associated_wallet,
        wallets=wallets,
        cookies=request.cookies,
    )


@devhelp_endpoint.route("/settings", methods=["POST"])
@login_required
@user_secret_decrypted_required
def settings_post():
    show_menu = request.form["show_menu"]
    user = app.specter.user_manager.get_user()
    if show_menu == "yes":
        user.add_service(DevhelpService.id)
    else:
        user.remove_service(DevhelpService.id)
    used_wallet_alias = request.form.get("used_wallet")
    if used_wallet_alias != None:
        wallet = current_user.wallet_manager.get_by_alias(used_wallet_alias)
        DevhelpService.set_associated_wallet(wallet)
    return redirect(url_for(f"{DevhelpService.get_blueprint_name()}.settings_get"))


NEVER_CALLED_PYTHON_COMMAND = True


@devhelp_endpoint.route("/python_command", methods=["POST"])
@login_required
def python_command():
    global NEVER_CALLED_PYTHON_COMMAND
    if NEVER_CALLED_PYTHON_COMMAND:
        NEVER_CALLED_PYTHON_COMMAND = False
        return robust_json_dumps(
            "!!!DANGER!!!\n"
            "This command allows arbitrary access to Specter.\n"
            "You can irreparabily damage your specter configuration and\n"
            "all funds on !!!HOT!!! wallets can be lost!\n"
            "Please use it with extreme care!!! Run your command again to execute it.\n\n"
            "--> Never copy&paste anything you do not FULLY understand. <--"
        )
    if current_user != "admin" or not app.specter.user_manager.user.is_admin:
        return robust_json_dumps(f"Access forbidden for user '{current_user}'!")
    if not app.config["DEVELOPER_JAVASCRIPT_PYTHON_CONSOLE"]:
        return robust_json_dumps(
            "DEVELOPER_JAVASCRIPT_PYTHON_CONSOLE disabled in Specter configuration.  "
            "This is an advanced option and should be used with great care!!!"
        )
    if request.method != "POST":
        return robust_json_dumps("Not a 'POST' command.")

    # The following commented lines are a further restriction of this endpoint, by limiting it only to regtest and testnet
    # uncomment these lines to enable the restriction
    # ----------------------------------------------
    # allowed_chains = ['regtest', 'testnet', 'liquidtestnet', 'liquidregtest']
    # if app.specter.chain not in allowed_chains:
    #     return robust_json_dumps(f"This command is only allowed for {allowed_chains}. "
    #                              "The current chain is {app.specter.chain}")
    # ----------------------------------------------

    command = request.form["command"]
    result = DevhelpService.console.exec_command(command)
    try:
        return robust_json_dumps(result)
    except:
        return robust_json_dumps(str(result))



import json
from json.decoder import JSONDecodeError
import logging
from math import isnan

import requests
from cryptoadvance.specter.specter_error import SpecterError
from cryptoadvance.specter.util.common import str2bool

from ..helpers import is_testnet, normalize_address
from ..util.descriptor import AddChecksum, Descriptor

logger = logging.getLogger(__name__)


class PsbtCreator:
    """A class to create PSBTs easily out of stuff coming from the frontend"""

    def __init__(
        self,
        specter,
        wallet,
        ui_option,
        request_form=None,
        recipients_txt=None,
        recipients_amount_unit=None,
        request_json=None,
    ):
        """
        * depending of ui_option = (ui|text) Fill the payment-details in either of these:
          * request_form: For details on the structure of the data for each recipient (amounts, addresses, etc.) see below at paymentinfo_from_ui
          * recipients_txt: expects the payment-details in textblock "recipients" and recipients_amount_unit for all
            amounts in recipients_txt either "sats" or "btc"
        * in both cases, the request_form also contains:
          * "substract": optional (default: False), Boolean whether to substract the fee from the amounts, otherwise additional input gets created
          * "subtract_from": index on which address to substract the fee from
          * fee_options:
            * "dynamic": get the fee from "fee_rate_dynamic"
            * "fee_rate": directly get the fee
          * "rbf": "on" or "off" "boolean" whether to replace-by-fee
          * "estimate_fee"
        """
        # Good to have some values for error-reporting
        self.ui_option = ui_option
        self.request_form = request_form
        self.request_json = request_json

        if ui_option == "ui":
            (
                self.addresses,
                self.labels,
                self.amounts,
                self.amount_units,
            ) = PsbtCreator.paymentinfo_from_ui(
                specter, wallet, request_form=request_form
            )
        elif ui_option == "text":
            if recipients_txt is None or recipients_amount_unit is None:
                raise SpecterError(
                    "recipients_txt and recipients_amount_unit is mandatory"
                )
            (
                self.addresses,
                self.labels,
                self.amounts,
                self.amount_units,
            ) = PsbtCreator.paymentinfo_from_text(
                specter,
                wallet,
                recipients_txt=recipients_txt,
                recipients_amount_unit=recipients_amount_unit,
            )
        elif ui_option == "json":
            if request_json is None:
                raise SpecterError("request_json is mandatory")
            (
                self.addresses,
                self.labels,
                self.amounts,
                self.amount_units,
            ) = PsbtCreator.paymentinfo_from_json(
                specter, wallet, request_json=request_json
            )
        else:
            raise SpecterError(
                f"Unknown ui_option: {ui_option}. Valid ones are ui|text|json"
            )
        # normalizing
        self.addresses = [normalize_address(addr) for addr in self.addresses]
        # get kwargs
        if ui_option == "ui" or ui_option == "text":
            self.kwargs = PsbtCreator.kwargs_from_request_form(request_form)
        elif ui_option == "json":
            self.kwargs = PsbtCreator.kwargs_from_request_json(request_json)
        if specter.is_liquid:
            self.kwargs["assets"] = self.amount_units
        self.validate_before_creation()

    def validate_before_creation(self):
        if self.kwargs["fee_rate"] == None:
            if self.ui_option == "ui" or self.ui_option == "text":
                additional_data = self.request_form
            elif self.ui_option == "json":
                additional_data = self.request_json

            raise Exception(
                f"Fee Rate could not be calculated and is now None. This is not supported right now and probably unintended anyway(ui_option = {self.ui_option}).\nrequest={additional_data}\nkwargs={self.kwargs}",
                additional_data,
            )

    def create_psbt(self, wallet):
        """creates the PSBT via the wallet and modifies it for if substract is true
        If there was a "estimate_fee" in the request_form, the PSBT will not get persisted
        """
        self.psbt = wallet.createpsbt(self.addresses, self.amounts, **self.kwargs)
        if self.psbt is None:
            raise SpecterError(
                "Probably you don't have enough funds, or something else..."
            )
        else:
            # calculate new amount if we need to subtract
            if self.kwargs["subtract"]:
                for v in self.psbt["tx"]["vout"]:
                    if self.addresses[0] in v["scriptPubKey"].get(
                        "addresses", [""]
                    ) or self.addresses[0] == v["scriptPubKey"].get("address", ""):
                        self.amounts[0] = v["value"]
        return self.psbt

    @classmethod
    def paymentinfo_from_ui(cls, specter, wallet, request_form):
        """Calculates the correct format needed by wallet.createpsbt() out of a request form.
        The recipient_dicts part in the form is a list of dicts and looks like this:
        [{'unit': 'btc', 'amount': 1, 'btc_amount': 1, 'recipient_id': 0, 'label': '', 'address': 'bcrt1q ... 58qwn'},
        {'unit': 'btc', 'amount': 2, 'btc_amount': 2, 'recipient_id': 1, 'label': '', 'address': 'bcrt1q ... vaa3p'},
        {'unit': 'btc', 'amount': 3, 'btc_amount': 3, 'recipient_id': 2, 'label': '', 'address': 'bcrt1q ... n0a85'}]

        Returns (addresses, labels, amounts, amount_units) (all arrays)
        """
        addresses = []
        labels = []
        amounts = []
        amount_units = []

        recipient_dicts = json.loads(request_form["recipient_dicts"])
        print(recipient_dicts)
        for recipient_dict in recipient_dicts:
            addresses.append(recipient_dict["address"])
            amount = 0.0
            try:
                amount = float(recipient_dict["btc_amount"])
            except ValueError:
                pass
            if isnan(amount):
                amount = 0.0
            amounts.append(amount)
            unit = recipient_dict["unit"]
            if specter.is_liquid and unit in ["sat", "btc"]:
                unit = specter.default_asset
            amount_units.append(unit)
            labels.append(recipient_dict["label"])
            if recipient_dict["label"] != "":
                wallet.setlabel(addresses[-1], labels[-1])

        return addresses, labels, amounts, amount_units

    @classmethod
    def paymentinfo_from_text(
        cls, specter, wallet, recipients_txt, recipients_amount_unit
    ):
        """calculates the correct format needed by wallet.createpsbt() out of a request-form
        out of a textbox holding addresses and amounts.
        """
        addresses = []
        labels = []
        amounts = []
        amount_units = []
        for output in recipients_txt.splitlines():
            try:
                if output.isspace() or output == "":
                    continue
                addresses.append(output.split(",")[0].strip())
                if recipients_amount_unit == "sat":
                    amounts.append(round(float(output.split(",")[1].strip()) / 1e8, 8))
                elif recipients_amount_unit == "btc":
                    amounts.append(float(output.split(",")[1].strip()))
                else:
                    raise SpecterError(
                        f"Unknown recipients_amount_unit: {recipients_amount_unit}"
                    )
                labels.append("")
                amount_units.append(recipients_amount_unit)
            except IndexError as ie:
                logger.error(f"line does not match expected pattern: '{output}'")
        return addresses, labels, amounts, amount_units

    @classmethod
    def paymentinfo_from_json(cls, specter, wallet, request_json):
        """calculates the correct format needed by wallet.createpsbt() out of a json
        Example:
        {
            "recipients" : [
                {
                    "address": "BCRT1qgc6h85z43g3ss2dl5zdrzrp3ef6av4neqcqhh8",
                    "amount": 0.1,
                    "unit": "btc",
                    "label": "someLabel"
                },
                {
                    "address": "bcrt1q3kfetuxpxvujasww6xas94nawklvpz0e52uw8a",
                    "amount": 111211,
                    "unit": "sat",
                    "label": "someOtherLabel"
                }
            ],
            "rbf_tx_id": "",
            "subtract_from": "1",
            "fee_rate": "64",
            "rbf": true,
        }
        returns something like  (addresses, labels, amounts, amount_units) (all arrays)
        """
        addresses = []
        labels = []
        amounts = []
        amount_units = []
        try:
            if isinstance(request_json, dict):
                json_data = request_json
            else:
                json_data = json.loads(request_json)

        except JSONDecodeError as e:
            raise SpecterError(f"Error parsing json: {e}")
        for recipient in json_data["recipients"]:
            try:
                addresses.append(recipient["address"])
                try:
                    amount = float(recipient["amount"])
                    if recipient["unit"] == "sat":
                        amounts.append(round(amount / 1e8, 8))
                    elif recipient["unit"] == "btc":
                        amounts.append(amount)
                    else:
                        if specter.is_liquid:
                            if len(recipient["unit"]) == 64:
                                amounts.append(amount)
                            else:
                                raise SpecterError(
                                    f"Non-compliant json: Unknown unit {recipient['unit']}. Accepted: sat, btc or 64-char hex asset id."
                                )
                        else:
                            raise SpecterError(
                                f"Non-compliant json: Unknown unit {recipient['unit']}. This could be caused by using a special name for your Elements regtest."
                            )
                except ValueError as e:
                    raise SpecterError(
                        f"Could not parse amount {recipient.get('amount')} because {e}"
                    )
                unit = recipient["unit"]
                if specter.is_liquid and unit in ["sat", "btc"]:
                    unit = specter.default_asset
                amount_units.append(unit)

                label = recipient.get("label", "")
                labels.append(label)
                if label != "":
                    wallet.setlabel(recipient["address"], label)
            except KeyError as ke:
                raise SpecterError(f"Data missing in json: {ke}")
        return addresses, labels, amounts, amount_units

    def kwargs_from_request_form(request_form):
        """calculates the needed kwargs fow wallet.createpsbt() out of a request_form"""
        # Who pays the fees?
        subtract = str2bool(request_form.get("subtract", False))
        subtract_from = int(request_form.get("subtract_from", 0))
        fee_option = request_form.get("fee_option")
        fee_rate = None
        if fee_option:
            if "dynamic" in fee_option:
                if request_form.get("fee_rate_dynamic"):
                    fee_rate = float(request_form.get("fee_rate_dynamic"))
                else:
                    raise Exception(
                        "fee_option is dynamic but no fee_rate_dynamic given",
                        request_form,
                    )
            else:
                if request_form.get("fee_rate"):
                    fee_rate = float(request_form.get("fee_rate"))
                else:
                    raise Exception(
                        "fee_option is manual but no fee_rate given", request_form
                    )
        rbf = str2bool(request_form.get("rbf", False))
        # workaround for making the tests work with a dict
        if hasattr(request_form, "getlist"):
            coins = [coin.split(",") for coin in request_form.getlist("coinselect")]
            # convert to tuples
            selected_coins = [
                {"txid": txid.strip(), "vout": int(vout)} for txid, vout in coins
            ]
        else:
            selected_coins = []
        rbf_tx_id = request_form.get("rbf_tx_id", "")
        kwargs = {
            "subtract": subtract,
            "subtract_from": subtract_from,
            "fee_rate": fee_rate,
            "rbf": rbf,
            "selected_coins": selected_coins,
            # determines whether the psbt gets persisted
            "readonly": "estimate_fee" in request_form,
            "rbf_edit_mode": (rbf_tx_id != ""),
        }
        return kwargs

    @classmethod
    def kwargs_from_request_json(cls, request_json):
        """calculates the needed kwargs fow wallet.createpsbt() out of a request_json"""
        # Who pays the fees?
        try:
            if isinstance(request_json, dict):
                json_data = request_json
            else:
                json_data = json.loads(request_json)

        except JSONDecodeError as e:
            raise SpecterError(f"Error parsing json: {e}")
        subtract = bool(json_data.get("subtract", False))
        subtract_from = int(json_data.get("subtract_from", 0))

        fee_rate = float(json_data.get("fee_rate", None))
        rbf = bool(json_data.get("rbf", False))
        rbf_tx_id = json_data.get("rbf_tx_id", "")
        kwargs = {
            "subtract": subtract,
            "subtract_from": subtract_from,
            "fee_rate": fee_rate,
            "rbf": rbf,
            "selected_coins": [],
            "readonly": False,  # determines whether the psbt gets persisted
            "rbf_edit_mode": (rbf_tx_id != ""),
        }
        return kwargs

    def __repr__(self) -> str:
        status = "created" if hasattr(self, "psbt") else "initialized"
        return f"<{self.__class__.__name__} amountSum={sum(self.amounts) } {status}>"



class Singleton:
    _instance = None

    def __init__(self):
        # Singleton pattern must prevent normal instantiation
        raise Exception(
            "Cannot directly instantiate a Singleton. Access via get_instance()"
        )

    @classmethod
    def get_instance(cls):
        # This is the only way to access the one and only instance
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
        return cls._instance


class ConfigurableSingletonException(Exception):
    pass


class ConfigurableSingleton(Singleton):
    @classmethod
    def get_instance(cls):
        # This is the only way to access the one and only instance
        if cls._instance:
            return cls._instance
        else:
            raise ConfigurableSingletonException(
                f"Must call {cls.__name__}.configure_instance(config) first"
            )

    @classmethod
    def configure_instance(cls, **kwargs):
        # Must be called before the first get_instance() call
        if cls._instance:
            raise ConfigurableSingletonException(f"{cls.__name__} already configured")

        # Instantiate the one and only instance
        cls._instance = cls.__new__(cls)



import logging
import urllib
from .sd_card_device import SDCardDevice
from ..util.xpub import get_xpub_fingerprint
from ..helpers import to_ascii20
from embit.psbt import PSBT, DerivationPath
from embit import bip32
from binascii import b2a_base64, a2b_base64
from collections import OrderedDict

logger = logging.getLogger(__name__)

CC_TYPES = {"legacy": "BIP45", "p2sh-segwit": "P2WSH-P2SH", "bech32": "P2WSH"}


class ColdCard(SDCardDevice):
    device_type = "coldcard"
    name = "ColdCard"
    icon = "coldcard_icon.svg"

    sd_card_support = True
    wallet_export_type = "file"
    supports_hwi_multisig_display_address = True

    def create_psbts(self, base64_psbt, wallet):
        psbts = super().create_psbts(base64_psbt, wallet)
        psbts["sdcard"] = wallet.fill_psbt(base64_psbt, non_witness=False, xpubs=True)
        return psbts

    def export_wallet(self, wallet):
        if not wallet.is_multisig:
            return None
        CC_TYPES = {"legacy": "BIP45", "p2sh-segwit": "P2WSH-P2SH", "bech32": "P2WSH"}
        # try to find at least one derivation
        # cc assume the same derivation for all keys :(
        derivation = None
        # find correct key
        for k in wallet.keys:
            if k in self.keys and k.derivation != "":
                derivation = k.derivation.replace("h", "'")
                break
        if derivation is None:
            return None
        cc_file = """# Coldcard Multisig setup file (created on Specter Desktop)
#
Name: {}
Policy: {} of {}
Derivation: {}
Format: {}
""".format(
            to_ascii20(wallet.name),
            wallet.sigs_required,
            len(wallet.keys),
            derivation,
            CC_TYPES[wallet.address_type],
        )
        for k in wallet.keys:
            # cc assumes fingerprint is known
            fingerprint = k.fingerprint
            if fingerprint == "":
                fingerprint = get_xpub_fingerprint(k.xpub).hex()
            cc_file += "{}: {}\n".format(fingerprint.upper(), k.xpub)
        return urllib.parse.quote(cc_file)



import datetime
import errno
import json
import logging
import os
import sys

import requests
import urllib3

from .helpers import is_ip_private
from .specter_error import SpecterError, handle_exception

logger = logging.getLogger(__name__)

# TODO: redefine __dir__ and help

RPC_PORTS = {
    "testnet": 18332,
    "test": 18332,
    "regtest": 18443,
    "main": 8332,
    "mainnet": 8332,
    "signet": 38332,
}


def get_default_datadir(node_type="BTC"):
    """Get default Bitcoin directory depending on the system"""
    datadir = None
    if node_type == "BTC":
        last_part = "Bitcoin"
    elif node_type == "ELM":
        last_part = "Elements"
    else:
        raise SpecterError(f"Unknown node_type {node_type}")

    if sys.platform == "darwin":
        # Not tested yet!
        datadir = os.path.join(
            os.environ["HOME"], f"Library/Application Support/{last_part}/"
        )
    elif sys.platform == "win32":
        # Not tested yet!
        datadir = os.path.join(os.environ["APPDATA"], last_part)
    else:
        datadir = os.path.join(os.environ["HOME"], f".{last_part.lower()}")
    return datadir


def _get_rpcconfig(datadir=get_default_datadir()):
    """returns the bitcoin.conf configurations (multiple) in a datastructure
    for all networks of a specific datadir.
    """
    config = {
        "bitcoin.conf": {"default": {}, "main": {}, "test": {}, "regtest": {}},
        "cookies": [],
    }
    if not os.path.isdir(datadir):  # we don't know where to search for files
        logger.warning(f"{datadir} not found")
        return config
    # load content from bitcoin.conf
    bitcoin_conf_file = os.path.join(datadir, "bitcoin.conf")
    if os.path.exists(bitcoin_conf_file):
        try:
            with open(bitcoin_conf_file, "r") as f:
                current = config["bitcoin.conf"]["default"]
                for line in f.readlines():
                    line = line.split("#")[0]

                    for net in config["bitcoin.conf"]:
                        if f"[{net}]" in line:
                            current = config["bitcoin.conf"][net]

                    if "=" not in line:
                        continue
                    k, v = line.split("=", 1)
                    # lines like main.rpcuser and so on
                    if "." in k:
                        net, k = k.split(".", 1)
                        config["bitcoin.conf"][net.strip()][k.strip()] = v.strip()
                    else:
                        current[k.strip()] = v.strip()
        except Exception as e:
            handle_exception(e)
            print("Can't open %s file" % bitcoin_conf_file)
    folders = {"main": "", "test": "testnet3", "regtest": "regtest", "signet": "signet"}
    for chain in folders:
        fname = os.path.join(datadir, folders[chain], ".cookie")
        if os.path.exists(fname):
            try:
                with open(fname, "r") as f:
                    content = f.read()
                    user, password = content.split(":")
                    obj = {"user": user, "password": password, "port": RPC_PORTS[chain]}
                    config["cookies"].append(obj)
            except Exception as e:
                handle_exception(e)
                print("Can't open %s file" % fname)
    return config


def _detect_rpc_confs_via_datadir(config=None, datadir=get_default_datadir()):
    if config is None:
        config = _get_rpcconfig(datadir=datadir)
    confs = []
    default = {}
    for network in config["bitcoin.conf"]:
        if "rpcuser" in config["bitcoin.conf"][network]:
            default["user"] = config["bitcoin.conf"][network]["rpcuser"]
        if "rpcpassword" in config["bitcoin.conf"][network]:
            default["password"] = config["bitcoin.conf"][network]["rpcpassword"]
        if "rpcconnect" in config["bitcoin.conf"][network]:
            default["host"] = config["bitcoin.conf"][network]["rpcconnect"]
        if "rpcport" in config["bitcoin.conf"][network]:
            default["port"] = int(config["bitcoin.conf"][network]["rpcport"])
        if "user" in default and "password" in default:
            if (
                "port" not in config["bitcoin.conf"]["default"]
            ):  # only one rpc makes sense in this case
                if network == "default":
                    continue
                default["port"] = RPC_PORTS[network]
            confs.append(default.copy())
    # try cookies now
    for cookie in config["cookies"]:
        o = {}
        o.update(default)
        o.update(cookie)
        confs.append(o)
    return confs


def _detect_rpc_confs_via_env(prefix):
    """returns an array which might contain one configmap derived from Env-Vars
    Env-Vars for prefix=BTC: BTC_RPC_USER, BTC_RPC_PASSWORD, BTC_RPC_HOST, BTC_RPC_PORT
    configmap: {"user":"user","password":"password","host":"host","port":"port","protocol":"https"}
    """
    rpc_arr = []
    if (
        os.getenv(f"{prefix}_RPC_USER")
        and os.getenv(f"{prefix}_RPC_PASSWORD")
        and os.getenv(f"{prefix}_RPC_HOST")
        and os.getenv(f"{prefix}_RPC_PORT")
        and os.getenv(f"{prefix}_RPC_PROTOCOL")
    ):
        logger.info(f"Detected RPC-Config on Environment-Variables for prefix {prefix}")
        env_conf = {
            "user": os.getenv(f"{prefix}_RPC_USER"),
            "password": os.getenv(f"{prefix}_RPC_PASSWORD"),
            "host": os.getenv(f"{prefix}_RPC_HOST"),
            "port": os.getenv(f"{prefix}_RPC_PORT"),
            "protocol": os.getenv(f"{prefix}_RPC_PROTOCOL"),
        }
        rpc_arr.append(env_conf)
    return rpc_arr


def autodetect_rpc_confs(
    node_type,
    datadir=get_default_datadir(),
    port=None,
    proxy_url="socks5h://localhost:9050",
    only_tor=False,
):
    """Returns an array of valid and working configurations which
    got autodetected.
    autodetection checks env-vars and bitcoin-data-dirs
    """
    if port == "":
        port = None
    if port is not None:
        port = int(port)
    conf_arr = []
    conf_arr.extend(_detect_rpc_confs_via_env(node_type))
    conf_arr.extend(_detect_rpc_confs_via_datadir(datadir=datadir))
    available_conf_arr = []
    if len(conf_arr) > 0:
        for conf in conf_arr:
            rpc = BitcoinRPC(
                **conf, proxy_url="socks5h://localhost:9050", only_tor=False
            )
            if port is not None:
                if int(rpc.port) != port:
                    continue
            try:
                rpc.getmininginfo()
                available_conf_arr.append(conf)
            except requests.exceptions.RequestException as e:
                pass
                # no point in reporting that here
            except RpcError:
                pass
                # have to make a list of acceptable exception unfortunately
                # please enlarge if you find new ones
    return available_conf_arr


class RpcError(Exception):
    """Specifically created for error-handling of the BitcoinCore-API
    if thrown, check for errors like this:
    try:
        rpc.does_not_exist()
    except RpcError as rpce:
        assert rpce.status_code == 401 # A https-status-code
        assert rpce.error_code == -32601
        assert rpce.error_msg == "Method not found"
    See for error_codes https://github.com/bitcoin/bitcoin/blob/v0.15.0.1/src/rpc/protocol.h#L32L87
    """

    def __init__(self, message, response):
        super(Exception, self).__init__(message)
        self.status_code = 500  # default
        try:
            self.status_code = response.status_code
            error = response.json()
        except Exception as e:
            # it's a dict already
            error = response
        try:
            self.error_code = error["error"]["code"]
            self.error_msg = error["error"]["message"]
        except Exception as e:
            self.error_code = -99
            self.error = "UNKNOWN API-ERROR:%s" % response.text


class BitcoinRPC:
    counter = 0

    # These are used for tracing the calls without too many duplicates
    last_call_hash = None
    last_call_hash_counter = 0

    # https://docs.python-requests.org/en/master/user/quickstart/#timeouts
    # None means until connection closes. It's specified in seconds
    default_timeout = None  # seconds

    def __init__(
        self,
        user="bitcoin",
        password="secret",
        host="127.0.0.1",
        port=8332,
        protocol="http",
        path="",
        timeout=None,
        session=None,
        proxy_url="socks5h://localhost:9050",
        only_tor=False,
        **kwargs,
    ):
        path = path.replace("//", "/")  # just in case
        self.user = user
        self._password = password
        self.port = port
        self.protocol = protocol
        self.host = host
        self.path = path
        self.timeout = timeout or self.__class__.default_timeout
        self.proxy_url = proxy_url
        self.only_tor = only_tor
        self.r = None
        self.last_call_hash = None
        self.last_call_hash_counter = 0
        # session reuse speeds up requests
        if session is None:
            self._create_session()
        else:
            self.session = session

    def _create_session(self):
        session = requests.Session()
        session.auth = (self.user, self.password)
        # check if we need to connect over Tor
        if not is_ip_private(self.host):
            if self.only_tor or ".onion" in self.host:
                # configure Tor proxies
                session.proxies["http"] = self.proxy_url
                session.proxies["https"] = self.proxy_url
        self.session = session

    def wallet(self, name=""):
        return type(self)(
            user=self.user,
            password=self.password,
            port=self.port,
            protocol=self.protocol,
            host=self.host,
            path="{}/wallet/{}".format(self.path, name),
            timeout=self.timeout,
            session=self.session,
            proxy_url=self.proxy_url,
            only_tor=self.only_tor,
        )

    @property
    def url(self):
        return "{s.protocol}://{s.host}:{s.port}{s.path}".format(s=self)

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value
        self._create_session()

    def test_connection(self):
        """returns a boolean depending on whether getblockchaininfo() succeeds"""
        try:
            self.getblockchaininfo()
            return True
        except:
            return False

    def clone(self):
        """
        Returns a clone of self.
        Useful if you want to mess with the properties
        """
        return BitcoinRPC(
            self.user,
            self.password,
            self.host,
            self.port,
            self.protocol,
            self.path,
            self.timeout,
            self.session,
            self.proxy_url,
            self.only_tor,
        )

    def multi(self, calls: list, **kwargs):
        """Makes batch request to Core"""
        type(self).counter += len(calls)
        # some debug info for optimizations
        # methods = " ".join(list(dict.fromkeys([call[0] for call in calls])))
        # wallet = self.path.split("/")[-1]
        # print(f"{self.counter}: +{len(calls)} {wallet} {methods}")
        headers = {"content-type": "application/json"}
        payload = [
            {
                "method": method,
                "params": args if args != [None] else [],
                "jsonrpc": "2.0",
                "id": i,
            }
            for i, (method, *args) in enumerate(calls)
        ]
        timeout = self.timeout
        if "timeout" in kwargs:
            timeout = kwargs["timeout"]

        if kwargs.get("no_wait"):
            # Zero is treated like None, i.e. infinite wait
            timeout = 0.001

        url = self.url
        if "wallet" in kwargs:
            url = url + "/wallet/{}".format(kwargs["wallet"])
        ts = self.trace_call_before(url, payload)
        try:
            r = self.session.post(
                url, data=json.dumps(payload), headers=headers, timeout=timeout
            )
        except (requests.exceptions.Timeout, urllib3.exceptions.ReadTimeoutError) as to:
            # Timeout is effectively one of the two:
            # ConnectTimeout: The request timed out while trying to connect to the remote server
            # ReadTimeout: The server did not send any data in the allotted amount of time.
            # ReadTimeoutError: Raised when a socket timeout occurs while receiving data from a server
            if kwargs.get("no_wait"):
                # Used for rpc calls that don't immediately return (e.g. rescanblockchain) so we don't
                # expect any data back anyway. __getattr__ expects a list of formatted json.
                self.trace_call_after(url, payload, timeout)
                return [{"error": None, "result": None}]

            logger.error(
                "Timeout after {} secs while {} call({: <28}) payload:{} Exception: {}".format(
                    timeout,
                    self.__class__.__name__,
                    "/".join(url.split("/")[3:]),
                    payload,
                    to,
                )
            )
            raise SpecterError(
                "Timeout after {} secs while {} call({: <28}). Check the logs for more details.".format(
                    timeout,
                    self.__class__.__name__,
                    "/".join(url.split("/")[3:]),
                    payload,
                )
            )
        self.trace_call_after(url, payload, ts)
        self.r = r
        if r.status_code != 200:
            logger.debug(f"last call FAILED: {r.text}")
            if r.text.startswith("Work queue depth exceeded"):
                raise SpecterError(
                    "Your Bitcoind is running hot (Work queue depth exceeded)! Bitcoind gets more requests than it can process. Please refrain from doing anything for some minutes."
                )
            raise RpcError(
                "Server responded with error code %d: %s" % (r.status_code, r.text), r
            )
        r = r.json()
        return r

    @classmethod
    def trace_call_before(cls, url, payload):
        """get a timestamp if needed in order to measure how long the call takes"""
        if logger.level == logging.DEBUG:
            return datetime.datetime.now()

    @classmethod
    def trace_call_after(cls, url, payload, timestamp):
        """logs out the call and its payload (if necessary), reduces noise by suppressing repeated calls"""
        if logger.level == logging.DEBUG:
            timediff_ms = int(
                (datetime.datetime.now() - timestamp).total_seconds() * 1000
            )
            current_hash = hash(
                json.dumps({"url": url, "payload": payload}, sort_keys=True)
            )
            if cls.last_call_hash == None:
                cls.last_call_hash = current_hash
                cls.last_call_hash_counter = 0
            elif cls.last_call_hash == current_hash:
                cls.last_call_hash_counter = cls.last_call_hash_counter + 1
                return
            else:
                if cls.last_call_hash_counter > 0:
                    logger.debug(f"call repeated {cls.last_call_hash_counter} times")
                    cls.last_call_hash_counter = 0
                    cls.last_call_hash = current_hash
                else:
                    cls.last_call_hash = current_hash
            logger.debug(
                "call({: <28})({: >5}ms)  payload:{}".format(
                    "/".join(url.split("/")[3:]), timediff_ms, payload
                )
            )

    def __getattr__(self, method):
        def fn(*args, **kwargs):
            r = self.multi([(method, *args)], **kwargs)[0]
            if r["error"] is not None:
                raise RpcError(
                    f"Request error for method {method}: {r['error']['message']}", r
                )
            return r["result"]

        return fn

    def __repr__(self) -> str:
        return f"<BitcoinRpc {self.url}>"


if __name__ == "__main__":

    rpc = BitcoinRPC(
        "bitcoinrpc", "foi3uf092ury97iufhjf30982hf928uew9jd209j", port=18443
    )

    print(rpc.url)

    print(rpc.getmininginfo())

    print(rpc.listwallets())

    ##### WORKING WITH WALLETS #########

    # print(rpc.getbalance(wallet=""))

    # or

    w = rpc.wallet("")  # will load default wallet.dat

    print(w.url)

    print(w.getbalance())  # now you can run -rpcwallet commands
